# Copyright 2012 Google Inc. All rights reserved.
# -*- coding: utf-8 -*-
"""
Nicely formatted cipher suite definitions for TLS

A list of cipher suites in the form of CipherSuite objects.
These are supposed to be immutable; don't mess with them.
"""


class CipherSuite(object):
    """
    Encapsulates a cipher suite.

    Members/args:
    * code: two-byte ID code, as int
    * kx: key exchange algorithm, e.g. 'RSA' or 'DHE'
    * auth: authentication algorithm, e.g. 'RSA' or 'DSS'
    * cipher: stream or block cipher algorithm, e.g. 'AES_128'
    * mode: mode of operation for block ciphers, e.g. 'CBC' or 'GCM'
    * mac: message authentication code algorithm, e.g. 'MD5' or 'SHA256'
    * name: cipher suite name as defined in the RFCs,
        e.g. 'TLS_RSA_WITH_RC4_40_MD5', can be generated by default from the
        other parameters
    * encoding: encoding algorithm, defaults to cipher+mode
    Additional members:
    * kx_auth: kx+auth algorithm, as 'KeyExchangeAlgorithm' in RFCs
    """

    def __init__(self, code, kx, auth, cipher, mode, mac, name=None, encoding=None):
        self.code = code
        # We strip trailing whitespace here because we want to format the
        # global table nicely while making pylint happy.
        self._kx = kx.rstrip()
        self._auth = auth.rstrip()
        self.cipher = cipher.rstrip()
        self.mode = mode.rstrip()
        self.mac = mac.rstrip()
        self._name = name
        self._encoding = encoding

    @property
    def kx(self):
        if self._kx == '':  # for PSK
            return self._auth
        else:
            return self._kx

    @property
    def auth(self):
        if self._auth == '':  # for RSA
            return self._kx
        else:
            return self._auth

    @property
    def kx_auth(self):
        if self._auth == '':  # for RSA
            return self._kx
        elif self._kx == '':  # for PSK
            return self._auth
        else:
            return self._kx + '_' + self._auth

    @property
    def encoding(self):
        if self._encoding is None:
            if self.mode == '':
                return self.cipher
            else:
                return self.cipher + '_' + self.mode
        else:
            return self._encoding

    @property
    def name(self):
        if self._name is None:
            if self.mac == '':  # for CCM and CCM_8 modes
                return 'TLS_' + self.kx_auth + '_WITH_' + self.encoding
            else:
                return 'TLS_' + self.kx_auth + '_WITH_' + self.encoding + '_' + self.mac
        else:
            return self._name

    def __repr__(self):
        return 'CipherSuite(0x%04x, %s)' % (self.code, self.name)

    MAC_SIZES = {
        'MD5': 16,
        'SHA': 20,
        'SHA256': 32,
        'SHA384': 48,
    }

    BLOCK_SIZES = {
        '3DES_EDE': 8,
        'AES_128': 16,
        'AES_256': 16,
        'ARIA': 16,
        'CAMELLIA_128': 16,
        'CAMELLIA_256': 16,
        'CHACHA20': 64,
        'DES': 8,
        'DES40': 8,
        'IDEA': 8,
        'RC2_40': 8,
        'RC4_40': None,
        'RC4_128': None,
        'SEED': 16,
    }

    @property
    def mac_size(self):
        """In bytes. Default to 0."""
        return self.MAC_SIZES.get(self.mac, 0)

    @property
    def block_size(self):
        """In bytes. Default to 1."""
        return self.BLOCK_SIZES.get(self.cipher, 1)

    @property
    def pfs(self):
        return self.kx in ('DHE', 'ECDHE')

    @property
    def aead(self):
        return self.mode in ('CCM', 'CCM_8', 'GCM')

    @property
    def anonymous(self):
        return self.auth.startswith('anon')


def get_unknown_ciphersuite(code):
    return CipherSuite(code, '', '', '', '', '', name='Unknown')


# master list of CipherSuite Objects
# Full list from IANA:
#   https://www.iana.org/assignments/tls-parameters/tls-parameters.xhtml
CIPHERSUITES = [
    # not a real cipher suite, can be ignored, see RFC5746
    CipherSuite(0x00ff, 'NULL', '        ', 'NULL    ', '    ', 'NULL', 'TLS_EMPTY_RENEGOTIATION_INFO'),
    # RFC7507
    CipherSuite(0x5600, '', '            ', '', '', '', 'TLS_FALLBACK'),
    CipherSuite(0xffff, '', '            ', '', '', '', 'UNKNOWN_CIPHER'),

    # RFC2246 : TLS 1.0
    CipherSuite(0x0000, 'NULL', '        ', 'NULL    ', '    ', 'NULL'),

    CipherSuite(0x0001, 'RSA', '         ', 'NULL    ', '    ', 'MD5'),
    CipherSuite(0x0002, 'RSA', '         ', 'NULL    ', '    ', 'SHA'),
    CipherSuite(0x0003, 'RSA_EXPORT', '  ', 'RC4_40  ', '    ', 'MD5'),
    CipherSuite(0x0004, 'RSA', '         ', 'RC4_128 ', '    ', 'MD5'),
    CipherSuite(0x0005, 'RSA', '         ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0x0006, 'RSA_EXPORT', '  ', 'RC2_40  ', 'CBC ', 'MD5', encoding='RC2_CBC_40'),
    CipherSuite(0x0007, 'RSA', '         ', 'IDEA    ', 'CBC ', 'SHA'),
    CipherSuite(0x0008, 'RSA_EXPORT', '  ', 'DES40   ', 'CBC ', 'SHA'),
    CipherSuite(0x0009, 'RSA', '         ', 'DES     ', 'CBC ', 'SHA'),
    CipherSuite(0x000a, 'RSA', '         ', '3DES_EDE', 'CBC ', 'SHA'),

    CipherSuite(0x000b, 'DH', 'DSS_EXPORT', 'DES40   ', 'CBC ', 'SHA'),
    CipherSuite(0x000c, 'DH', 'DSS       ', 'DES     ', 'CBC ', 'SHA'),
    CipherSuite(0x000d, 'DH', 'DSS       ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0x000e, 'DH', 'RSA_EXPORT', 'DES40   ', 'CBC ', 'SHA'),
    CipherSuite(0x000f, 'DH', 'RSA       ', 'DES     ', 'CBC ', 'SHA'),
    CipherSuite(0x0010, 'DH', 'RSA       ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0x0011, 'DHE', 'DSS_EXPORT', 'DES40  ', 'CBC ', 'SHA'),
    CipherSuite(0x0012, 'DHE', 'DSS      ', 'DES     ', 'CBC ', 'SHA'),
    CipherSuite(0x0013, 'DHE', 'DSS      ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0x0014, 'DHE', 'RSA_EXPORT', 'DES40  ', 'CBC ', 'SHA'),
    CipherSuite(0x0015, 'DHE', 'RSA      ', 'DES     ', 'CBC ', 'SHA'),
    CipherSuite(0x0016, 'DHE', 'RSA      ', '3DES_EDE', 'CBC ', 'SHA'),

    CipherSuite(0x0017, 'DH', 'anon_EXPORT', 'RC4_40 ', '    ', 'MD5'),
    CipherSuite(0x0018, 'DH', 'anon      ', 'RC4_128 ', '    ', 'MD5'),
    CipherSuite(0x0019, 'DH', 'anon_EXPORT', 'DES40  ', 'CBC ', 'SHA'),
    CipherSuite(0x001a, 'DH', 'anon      ', 'DES     ', 'CBC ', 'SHA'),
    CipherSuite(0x001b, 'DH', 'anon      ', '3DES_EDE', 'CBC ', 'SHA'),

    # Reserved: 0x1c-0x1d

    # RFC4346 : TLS 1.1
    # RFC2712
    CipherSuite(0x001e, 'KRB5', '        ', 'DES     ', 'CBC ', 'SHA'),
    CipherSuite(0x001f, 'KRB5', '        ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0x0020, 'KRB5', '        ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0x0021, 'KRB5', '        ', 'IDEA    ', 'CBC ', 'SHA'),
    CipherSuite(0x0022, 'KRB5', '        ', 'DES     ', 'CBC ', 'MD5'),
    CipherSuite(0x0023, 'KRB5', '        ', '3DES_EDE', 'CBC ', 'MD5'),
    CipherSuite(0x0024, 'KRB5', '        ', 'RC4_128 ', '    ', 'MD5'),
    CipherSuite(0x0025, 'KRB5', '        ', 'IDEA    ', 'CBC ', 'MD5'),

    CipherSuite(0x0026, 'KRB5_EXPORT', ' ', 'DES40   ', 'CBC ', 'SHA', encoding='DES_CBC_40'),
    CipherSuite(0x0027, 'KRB5_EXPORT', ' ', 'RC2_40  ', 'CBC ', 'SHA', encoding='RC2_CBC_40'),
    CipherSuite(0x0028, 'KRB5_EXPORT', ' ', 'RC4_40  ', '    ', 'SHA'),
    CipherSuite(0x0029, 'KRB5_EXPORT', ' ', 'DES40   ', 'CBC ', 'MD5', encoding='DES_CBC_40'),
    CipherSuite(0x002a, 'KRB5_EXPORT', ' ', 'RC2_40  ', 'CBC ', 'MD5', encoding='RC2_CBC_40'),
    CipherSuite(0x002b, 'KRB5_EXPORT', ' ', 'RC4_40  ', '    ', 'MD5'),

    # RFC4785
    CipherSuite(0x002c, '    ', 'PSK     ', 'NULL    ', '    ', 'SHA'),
    CipherSuite(0x002d, 'DHE ', 'PSK     ', 'NULL    ', '    ', 'SHA'),
    CipherSuite(0x002e, 'RSA ', 'PSK     ', 'NULL    ', '    ', 'SHA'),

    # RFC3268
    CipherSuite(0x002f, 'RSA ', '        ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0x0030, 'DH  ', 'DSS     ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0x0031, 'DH  ', 'RSA     ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0x0032, 'DHE ', 'DSS     ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0x0033, 'DHE ', 'RSA     ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0x0034, 'DH  ', 'anon    ', 'AES_128 ', 'CBC ', 'SHA'),

    CipherSuite(0x0035, 'RSA ', '        ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0x0036, 'DH  ', 'DSS     ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0x0037, 'DH  ', 'RSA     ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0x0038, 'DHE ', 'DSS     ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0x0039, 'DHE ', 'RSA     ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0x003a, 'DH  ', 'anon    ', 'AES_256 ', 'CBC ', 'SHA'),

    # RFC5246 : TLS 1.2
    CipherSuite(0x003b, 'RSA ', '        ', 'NULL    ', '    ', 'SHA256'),
    CipherSuite(0x003c, 'RSA ', '        ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0x003d, 'RSA ', '        ', 'AES_256 ', 'CBC ', 'SHA256'),
    CipherSuite(0x003e, 'DH  ', 'DSS     ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0x003f, 'DH  ', 'RSA     ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0x0040, 'DHE ', 'DSS     ', 'AES_128 ', 'CBC ', 'SHA256'),

    # RFC5932
    CipherSuite(0x0041, 'RSA ', '        ', 'CAMELLIA_128', 'CBC', 'SHA'),
    CipherSuite(0x0042, 'DH  ', 'DSS     ', 'CAMELLIA_128', 'CBC', 'SHA'),
    CipherSuite(0x0043, 'DH  ', 'RSA     ', 'CAMELLIA_128', 'CBC', 'SHA'),
    CipherSuite(0x0044, 'DHE ', 'DSS     ', 'CAMELLIA_128', 'CBC', 'SHA'),
    CipherSuite(0x0045, 'DHE ', 'RSA     ', 'CAMELLIA_128', 'CBC', 'SHA'),
    CipherSuite(0x0046, 'DH  ', 'anon    ', 'CAMELLIA_128', 'CBC', 'SHA'),

    # Reserved: 0x47-5c
    # Unassigned: 0x5d-5f
    # Reserved: 0x60-66

    # RFC5246 : TLS 1.2
    CipherSuite(0x0067, 'DHE ', 'RSA     ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0x0068, 'DH  ', 'DSS     ', 'AES_256 ', 'CBC ', 'SHA256'),
    CipherSuite(0x0069, 'DH  ', 'RSA     ', 'AES_256 ', 'CBC ', 'SHA256'),
    CipherSuite(0x006a, 'DHE ', 'DSS     ', 'AES_256 ', 'CBC ', 'SHA256'),
    CipherSuite(0x006b, 'DHE ', 'RSA     ', 'AES_256 ', 'CBC ', 'SHA256'),
    CipherSuite(0x006c, 'DH  ', 'anon    ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0x006d, 'DH  ', 'anon    ', 'AES_256 ', 'CBC ', 'SHA256'),

    # Unassigned: 0x6e-83

    # RFC5932
    CipherSuite(0x0084, 'RSA ', '        ', 'CAMELLIA_256', 'CBC', 'SHA'),
    CipherSuite(0x0085, 'DH  ', 'DSS     ', 'CAMELLIA_256', 'CBC', 'SHA'),
    CipherSuite(0x0086, 'DH  ', 'RSA     ', 'CAMELLIA_256', 'CBC', 'SHA'),
    CipherSuite(0x0087, 'DHE ', 'DSS     ', 'CAMELLIA_256', 'CBC', 'SHA'),
    CipherSuite(0x0088, 'DHE ', 'RSA     ', 'CAMELLIA_256', 'CBC', 'SHA'),
    CipherSuite(0x0089, 'DH  ', 'anon    ', 'CAMELLIA_256', 'CBC', 'SHA'),

    # RFC4279
    CipherSuite(0x008a, '    ', 'PSK     ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0x008b, '    ', 'PSK     ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0x008c, '    ', 'PSK     ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0x008d, '    ', 'PSK     ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0x008e, 'DHE ', 'PSK     ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0x008f, 'DHE ', 'PSK     ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0x0090, 'DHE ', 'PSK     ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0x0091, 'DHE ', 'PSK     ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0x0092, 'RSA ', 'PSK     ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0x0093, 'RSA ', 'PSK     ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0x0094, 'RSA ', 'PSK     ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0x0095, 'RSA ', 'PSK     ', 'AES_256 ', 'CBC ', 'SHA'),

    # RFC4162
    CipherSuite(0x0096, 'RSA ', '        ', 'SEED    ', 'CBC ', 'SHA'),
    CipherSuite(0x0097, 'DH  ', 'DSS     ', 'SEED    ', 'CBC ', 'SHA'),
    CipherSuite(0x0098, 'DH  ', 'RSA     ', 'SEED    ', 'CBC ', 'SHA'),
    CipherSuite(0x0099, 'DHE ', 'DSS     ', 'SEED    ', 'CBC ', 'SHA'),
    CipherSuite(0x009a, 'DHE ', 'RSA     ', 'SEED    ', 'CBC ', 'SHA'),
    CipherSuite(0x009b, 'DH  ', 'anon    ', 'SEED    ', 'CBC ', 'SHA'),

    # RFC5288
    CipherSuite(0x009c, 'RSA ', '        ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0x009d, 'RSA ', '        ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0x009e, 'DHE ', 'RSA     ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0x009f, 'DHE ', 'RSA     ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0x00a0, 'DH  ', 'RSA     ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0x00a1, 'DH  ', 'RSA     ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0x00a2, 'DHE ', 'DSS     ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0x00a3, 'DHE ', 'DSS     ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0x00a4, 'DH  ', 'DSS     ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0x00a5, 'DH  ', 'DSS     ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0x00a6, 'DH  ', 'anon    ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0x00a7, 'DH  ', 'anon    ', 'AES_256 ', 'GCM ', 'SHA384'),

    # RFC5487
    CipherSuite(0x00a8, '    ', 'PSK     ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0x00a9, '    ', 'PSK     ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0x00aa, 'DHE ', 'PSK     ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0x00ab, 'DHE ', 'PSK     ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0x00ac, 'RSA ', 'PSK     ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0x00ad, 'RSA ', 'PSK     ', 'AES_256 ', 'GCM ', 'SHA384'),

    CipherSuite(0x00ae, '    ', 'PSK     ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0x00af, '    ', 'PSK     ', 'AES_256 ', 'CBC ', 'SHA384'),
    CipherSuite(0x00b0, '    ', 'PSK     ', 'NULL    ', '    ', 'SHA256'),
    CipherSuite(0x00b1, '    ', 'PSK     ', 'NULL    ', '    ', 'SHA384'),

    CipherSuite(0x00b2, 'DHE ', 'PSK     ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0x00b3, 'DHE ', 'PSK     ', 'AES_256 ', 'CBC ', 'SHA384'),
    CipherSuite(0x00b4, 'DHE ', 'PSK     ', 'NULL    ', '    ', 'SHA256'),
    CipherSuite(0x00b5, 'DHE ', 'PSK     ', 'NULL    ', '    ', 'SHA384'),

    CipherSuite(0x00b6, 'RSA ', 'PSK     ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0x00b7, 'RSA ', 'PSK     ', 'AES_256 ', 'CBC ', 'SHA384'),
    CipherSuite(0x00b8, 'RSA ', 'PSK     ', 'NULL    ', '    ', 'SHA256'),
    CipherSuite(0x00b9, 'RSA ', 'PSK     ', 'NULL    ', '    ', 'SHA384'),

    # RFC5932
    CipherSuite(0x00ba, 'RSA ', '        ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0x00bb, 'DH  ', 'DSS     ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0x00bc, 'DH  ', 'RSA     ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0x00bd, 'DHE ', 'DSS     ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0x00be, 'DHE ', 'RSA     ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0x00bf, 'DH  ', 'anon    ', 'CAMELLIA_128', 'CBC', 'SHA256'),

    CipherSuite(0x00c0, 'RSA ', '        ', 'CAMELLIA_256', 'CBC', 'SHA256'),
    CipherSuite(0x00c1, 'DH  ', 'DSS     ', 'CAMELLIA_256', 'CBC', 'SHA256'),
    CipherSuite(0x00c2, 'DH  ', 'RSA     ', 'CAMELLIA_256', 'CBC', 'SHA256'),
    CipherSuite(0x00c3, 'DHE ', 'DSS     ', 'CAMELLIA_256', 'CBC', 'SHA256'),
    CipherSuite(0x00c4, 'DHE ', 'RSA     ', 'CAMELLIA_256', 'CBC', 'SHA256'),
    CipherSuite(0x00c5, 'DH  ', 'anon    ', 'CAMELLIA_256', 'CBC', 'SHA256'),

    # RFC8446 TLS 1.3
    CipherSuite(0x1301, '    ', '        ', 'AES_128 ', 'GCM     ', 'SHA256', name='TLS_AES_128_GCM_SHA256'),
    CipherSuite(0x1302, '    ', '        ', 'AES_256 ', 'GCM     ', 'SHA384', name='TLS_AES_256_GCM_SHA384'),
    CipherSuite(0x1303, '    ', '        ', 'CHACHA20', 'POLY1305', 'SHA256', name='TLS_CHACHA20_POLY1305_SHA256'),
    CipherSuite(0x1304, '    ', '        ', 'AES_128 ', 'CCM     ', 'SHA256', name='TLS_AES_128_CCM_SHA256'),
    CipherSuite(0x1305, '    ', '        ', 'AES_128 ', 'CCM_8   ', 'SHA256', name='TLS_AES_128_CCM_8_SHA256'),

    # RFC4492
    CipherSuite(0xc001, 'ECDH ', 'ECDSA  ', 'NULL    ', '    ', 'SHA'),
    CipherSuite(0xc002, 'ECDH ', 'ECDSA  ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0xc003, 'ECDH ', 'ECDSA  ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0xc004, 'ECDH ', 'ECDSA  ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0xc005, 'ECDH ', 'ECDSA  ', 'AES_256 ', 'CBC ', 'SHA'),

    CipherSuite(0xc006, 'ECDHE', 'ECDSA  ', 'NULL    ', '    ', 'SHA'),
    CipherSuite(0xc007, 'ECDHE', 'ECDSA  ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0xc008, 'ECDHE', 'ECDSA  ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0xc009, 'ECDHE', 'ECDSA  ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0xc00a, 'ECDHE', 'ECDSA  ', 'AES_256 ', 'CBC ', 'SHA'),

    CipherSuite(0xc00b, 'ECDH ', 'RSA    ', 'NULL    ', '    ', 'SHA'),
    CipherSuite(0xc00c, 'ECDH ', 'RSA    ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0xc00d, 'ECDH ', 'RSA    ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0xc00e, 'ECDH ', 'RSA    ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0xc00f, 'ECDH ', 'RSA    ', 'AES_256 ', 'CBC ', 'SHA'),

    CipherSuite(0xc010, 'ECDHE', 'RSA    ', 'NULL    ', '    ', 'SHA'),
    CipherSuite(0xc011, 'ECDHE', 'RSA    ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0xc012, 'ECDHE', 'RSA    ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0xc013, 'ECDHE', 'RSA    ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0xc014, 'ECDHE', 'RSA    ', 'AES_256 ', 'CBC ', 'SHA'),

    CipherSuite(0xc015, 'ECDH ', 'anon   ', 'NULL    ', '    ', 'SHA'),
    CipherSuite(0xc016, 'ECDH ', 'anon   ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0xc017, 'ECDH ', 'anon   ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0xc018, 'ECDH ', 'anon   ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0xc019, 'ECDH ', 'anon   ', 'AES_256 ', 'CBC ', 'SHA'),

    # RFC5054
    CipherSuite(0xc01a, 'SRP_SHA', '     ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0xc01b, 'SRP_SHA', 'RSA  ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0xc01c, 'SRP_SHA', 'DSS  ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0xc01d, 'SRP_SHA', '     ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0xc01e, 'SRP_SHA', 'RSA  ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0xc01f, 'SRP_SHA', 'DSS  ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0xc020, 'SRP_SHA', '     ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0xc021, 'SRP_SHA', 'RSA  ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0xc022, 'SRP_SHA', 'DSS  ', 'AES_256 ', 'CBC ', 'SHA'),

    # RFC5289
    CipherSuite(0xc023, 'ECDHE', 'ECDSA  ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0xc024, 'ECDHE', 'ECDSA  ', 'AES_256 ', 'CBC ', 'SHA384'),
    CipherSuite(0xc025, 'ECDH ', 'ECDSA  ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0xc026, 'ECDH ', 'ECDSA  ', 'AES_256 ', 'CBC ', 'SHA384'),
    CipherSuite(0xc027, 'ECDHE', 'RSA    ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0xc028, 'ECDHE', 'RSA    ', 'AES_256 ', 'CBC ', 'SHA384'),
    CipherSuite(0xc029, 'ECDH ', 'RSA    ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0xc02a, 'ECDH ', 'RSA    ', 'AES_256 ', 'CBC ', 'SHA384'),

    CipherSuite(0xc02b, 'ECDHE', 'ECDSA  ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0xc02c, 'ECDHE', 'ECDSA  ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0xc02d, 'ECDH ', 'ECDSA  ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0xc02e, 'ECDH ', 'ECDSA  ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0xc02f, 'ECDHE', 'RSA    ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0xc030, 'ECDHE', 'RSA    ', 'AES_256 ', 'GCM ', 'SHA384'),
    CipherSuite(0xc031, 'ECDH ', 'RSA    ', 'AES_128 ', 'GCM ', 'SHA256'),
    CipherSuite(0xc032, 'ECDH ', 'RSA    ', 'AES_256 ', 'GCM ', 'SHA384'),

    # RFC5489
    CipherSuite(0xc033, 'ECDHE', 'PSK    ', 'RC4_128 ', '    ', 'SHA'),
    CipherSuite(0xc034, 'ECDHE', 'PSK    ', '3DES_EDE', 'CBC ', 'SHA'),
    CipherSuite(0xc035, 'ECDHE', 'PSK    ', 'AES_128 ', 'CBC ', 'SHA'),
    CipherSuite(0xc036, 'ECDHE', 'PSK    ', 'AES_256 ', 'CBC ', 'SHA'),
    CipherSuite(0xc037, 'ECDHE', 'PSK    ', 'AES_128 ', 'CBC ', 'SHA256'),
    CipherSuite(0xc038, 'ECDHE', 'PSK    ', 'AES_256 ', 'CBC ', 'SHA384'),
    CipherSuite(0xc039, 'ECDHE', 'PSK    ', 'NULL    ', '    ', 'SHA'),
    CipherSuite(0xc03a, 'ECDHE', 'PSK    ', 'NULL    ', '    ', 'SHA256'),
    CipherSuite(0xc03b, 'ECDHE', 'PSK    ', 'NULL    ', '    ', 'SHA384'),

    # RFC6209
    CipherSuite(0xc03c, 'RSA ', '        ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc03d, 'RSA ', '        ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc03e, 'DH  ', 'DSS     ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc03f, 'DH  ', 'DSS     ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc040, 'DH  ', 'RSA     ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc041, 'DH  ', 'RSA     ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc042, 'DHE ', 'DSS     ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc043, 'DHE ', 'DSS     ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc044, 'DHE ', 'RSA     ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc045, 'DHE ', 'RSA     ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc046, 'DH  ', 'anon    ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc047, 'DH  ', 'anon    ', 'ARIA_256', 'CBC ', 'SHA384'),

    CipherSuite(0xc048, 'ECDHE', 'ECDSA  ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc049, 'ECDHE', 'ECDSA  ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc04a, 'ECDH ', 'ECDSA  ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc04b, 'ECDH ', 'ECDSA  ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc04c, 'ECDHE', 'RSA    ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc04d, 'ECDHE', 'RSA    ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc04e, 'ECDH ', 'RSA    ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc04f, 'ECDH ', 'RSA    ', 'ARIA_256', 'CBC ', 'SHA384'),

    CipherSuite(0xc050, 'RSA ', '        ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc051, 'RSA ', '        ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc052, 'DHE ', 'RSA     ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc053, 'DHE ', 'RSA     ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc054, 'DH  ', 'RSA     ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc055, 'DH  ', 'RSA     ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc056, 'DHE ', 'DSS     ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc057, 'DHE ', 'DSS     ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc058, 'DH  ', 'DSS     ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc059, 'DH  ', 'DSS     ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc05a, 'DH  ', 'anon    ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc05b, 'DH  ', 'anon    ', 'ARIA_256', 'GCM ', 'SHA384'),

    CipherSuite(0xc05c, 'ECDHE', 'ECDSA  ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc05d, 'ECDHE', 'ECDSA  ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc05e, 'ECDH ', 'ECDSA  ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc05f, 'ECDH ', 'ECDSA  ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc060, 'ECDHE', 'RSA    ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc061, 'ECDHE', 'RSA    ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc062, 'ECDH ', 'RSA    ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc063, 'ECDH ', 'RSA    ', 'ARIA_256', 'GCM ', 'SHA384'),

    CipherSuite(0xc064, '    ', 'PSK     ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc065, '    ', 'PSK     ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc066, 'DHE ', 'PSK     ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc067, 'DHE ', 'PSK     ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc068, 'RSA ', 'PSK     ', 'ARIA_128', 'CBC ', 'SHA256'),
    CipherSuite(0xc069, 'RSA ', 'PSK     ', 'ARIA_256', 'CBC ', 'SHA384'),
    CipherSuite(0xc06a, '    ', 'PSK     ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc06b, '    ', 'PSK     ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc06c, 'DHE ', 'PSK     ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc06d, 'DHE ', 'PSK     ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc06e, 'RSA ', 'PSK     ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc06f, 'RSA ', 'PSK     ', 'ARIA_256', 'GCM ', 'SHA384'),
    CipherSuite(0xc070, 'ECDHE', 'PSK    ', 'ARIA_128', 'GCM ', 'SHA256'),
    CipherSuite(0xc071, 'ECDHE', 'PSK    ', 'ARIA_256', 'GCM ', 'SHA384'),

    # RFC6367
    CipherSuite(0xc072, 'ECDHE', 'ECDSA  ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0xc073, 'ECDHE', 'ECDSA  ', 'CAMELLIA_256', 'CBC', 'SHA384'),
    CipherSuite(0xc074, 'ECDH ', 'ECDSA  ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0xc075, 'ECDH ', 'ECDSA  ', 'CAMELLIA_256', 'CBC', 'SHA384'),
    CipherSuite(0xc076, 'ECDHE', 'RSA    ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0xc077, 'ECDHE', 'RSA    ', 'CAMELLIA_256', 'CBC', 'SHA384'),
    CipherSuite(0xc078, 'ECDH ', 'RSA    ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0xc079, 'ECDH ', 'RSA    ', 'CAMELLIA_256', 'CBC', 'SHA384'),

    CipherSuite(0xc07a, 'RSA ', '        ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc07b, 'RSA ', '        ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc07c, 'DHE ', 'RSA     ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc07d, 'DHE ', 'RSA     ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc07e, 'DH  ', 'RSA     ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc07f, 'DH  ', 'RSA     ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc080, 'DHE ', 'DSS     ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc081, 'DHE ', 'DSS     ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc082, 'DH  ', 'DSS     ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc083, 'DH  ', 'DSS     ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc084, 'DH  ', 'anon    ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc085, 'DH  ', 'anon    ', 'CAMELLIA_256', 'GCM', 'SHA384'),

    CipherSuite(0xc086, 'ECDHE', 'ECDSA  ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc087, 'ECDHE', 'ECDSA  ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc088, 'ECDH ', 'ECDSA  ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc089, 'ECDH ', 'ECDSA  ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc08a, 'ECDHE', 'RSA    ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc08b, 'ECDHE', 'RSA    ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc08c, 'ECDH ', 'RSA    ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc08d, 'ECDH ', 'RSA    ', 'CAMELLIA_256', 'GCM', 'SHA384'),

    CipherSuite(0xc08e, '    ', 'PSK     ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc08f, '    ', 'PSK     ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc090, 'DHE ', 'PSK     ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc091, 'DHE ', 'PSK     ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc092, 'RSA ', 'PSK     ', 'CAMELLIA_128', 'GCM', 'SHA256'),
    CipherSuite(0xc093, 'RSA ', 'PSK     ', 'CAMELLIA_256', 'GCM', 'SHA384'),
    CipherSuite(0xc094, '    ', 'PSK     ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0xc095, '    ', 'PSK     ', 'CAMELLIA_256', 'CBC', 'SHA384'),
    CipherSuite(0xc096, 'DHE ', 'PSK     ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0xc097, 'DHE ', 'PSK     ', 'CAMELLIA_256', 'CBC', 'SHA384'),
    CipherSuite(0xc098, 'RSA ', 'PSK     ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0xc099, 'RSA ', 'PSK     ', 'CAMELLIA_256', 'CBC', 'SHA384'),
    CipherSuite(0xc09a, 'ECDHE', 'PSK    ', 'CAMELLIA_128', 'CBC', 'SHA256'),
    CipherSuite(0xc09b, 'ECDHE', 'PSK    ', 'CAMELLIA_256', 'CBC', 'SHA384'),

    # RFC6655
    CipherSuite(0xc09c, 'RSA ', '        ', 'AES_128 ', 'CCM ', ''),
    CipherSuite(0xc09d, 'RSA ', '        ', 'AES_256 ', 'CCM ', ''),
    CipherSuite(0xc09e, 'DHE ', 'RSA     ', 'AES_128 ', 'CCM ', ''),
    CipherSuite(0xc09f, 'DHE ', 'RSA     ', 'AES_256 ', 'CCM ', ''),
    CipherSuite(0xc0a0, 'RSA ', '        ', 'AES_128 ', 'CCM_8', ''),
    CipherSuite(0xc0a1, 'RSA ', '        ', 'AES_256 ', 'CCM_8', ''),
    CipherSuite(0xc0a2, 'DHE ', 'RSA     ', 'AES_128 ', 'CCM_8', ''),
    CipherSuite(0xc0a3, 'DHE ', 'RSA     ', 'AES_256 ', 'CCM_8', ''),

    CipherSuite(0xc0a4, '    ', 'PSK     ', 'AES_128 ', 'CCM ', ''),
    CipherSuite(0xc0a5, '    ', 'PSK     ', 'AES_256 ', 'CCM ', ''),
    CipherSuite(0xc0a6, 'DHE ', 'PSK     ', 'AES_128 ', 'CCM ', ''),
    CipherSuite(0xc0a7, 'DHE ', 'PSK     ', 'AES_256 ', 'CCM ', ''),
    CipherSuite(0xc0a8, '    ', 'PSK     ', 'AES_128 ', 'CCM_8', ''),
    CipherSuite(0xc0a9, '    ', 'PSK     ', 'AES_256 ', 'CCM_8', ''),
    CipherSuite(0xc0aa, 'DHE ', 'PSK     ', 'AES_128 ', 'CCM_8', ''),
    CipherSuite(0xc0ab, 'DHE ', 'PSK     ', 'AES_256 ', 'CCM_8', ''),

    # RFC7251
    CipherSuite(0xc0ac, 'ECDHE', 'ECDSA  ', 'AES_128 ', 'CCM ', ''),
    CipherSuite(0xc0ad, 'ECDHE', 'ECDSA  ', 'AES_256 ', 'CCM ', ''),
    CipherSuite(0xc0ae, 'ECDHE', 'ECDSA  ', 'AES_128 ', 'CCM_8', ''),
    CipherSuite(0xc0af, 'ECDHE', 'ECDSA  ', 'AES_256 ', 'CCM_8', ''),

    # Unassigned: 0xc0b0-0xcca7
    CipherSuite(0xcc13, 'ECDHE', 'RSA    ', 'CHACHA20', 'POLY1305', 'SHA256',
                'OLD_TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256'),
    CipherSuite(0xcc14, 'ECDHE', 'ECDSA  ', 'CHACHA20', 'POLY1305', 'SHA256',
                'OLD_TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256'),
    CipherSuite(0xcc15, 'DHE  ', 'RSA    ', 'CHACHA20', 'POLY1305', 'SHA256',
                'OLD_TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256'),

    # RFC7905
    CipherSuite(0xcca8, 'ECDHE', 'RSA    ', 'CHACHA20', 'POLY1305', 'SHA256'),
    CipherSuite(0xcca9, 'ECDHE', 'ECDSA  ', 'CHACHA20', 'POLY1305', 'SHA256'),
    CipherSuite(0xccaa, 'DHE  ', 'RSA    ', 'CHACHA20', 'POLY1305', 'SHA256'),

    CipherSuite(0xccab, '     ', 'PSK    ', 'CHACHA20', 'POLY1305', 'SHA256'),
    CipherSuite(0xccac, 'ECDHE', 'PSK    ', 'CHACHA20', 'POLY1305', 'SHA256'),
    CipherSuite(0xccad, 'DHE  ', 'PSK    ', 'CHACHA20', 'POLY1305', 'SHA256'),
    CipherSuite(0xccae, 'RSA  ', 'PSK    ', 'CHACHA20', 'POLY1305', 'SHA256'),

    # RFC8701  // GREASE (Generate Random Extensions And Sustain Extensibility)
    CipherSuite(0x0a0a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0x1a1a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0x2a2a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0x3a3a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0x4a4a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0x5a5a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0x6a6a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0x7a7a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0x8a8a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0x9a9a, '', '', '', '', '', 'GREASE'),
    CipherSuite(0xaaaa, '', '', '', '', '', 'GREASE'),
    CipherSuite(0xbaba, '', '', '', '', '', 'GREASE'),
    CipherSuite(0xcaca, '', '', '', '', '', 'GREASE'),
    CipherSuite(0xdada, '', '', '', '', '', 'GREASE'),
    CipherSuite(0xeaea, '', '', '', '', '', 'GREASE'),
    CipherSuite(0xfafa, '', '', '', '', '', 'GREASE'),

    # Unassigned: 0xccaf-0xfefd
    # Reserved: 0xfefe-0xffff

]

BY_CODE = dict(
    (cipher.code, cipher) for cipher in CIPHERSUITES)

# This is a function to avoid artificially increased coverage
BY_NAME_DICT = None


def BY_NAME(name):
    # We initialize the dictionary only on the first call
    global BY_NAME_DICT
    if BY_NAME_DICT is None:
        BY_NAME_DICT = dict((suite.name, suite) for suite in CIPHERSUITES)
    return BY_NAME_DICT[name]


NULL_SUITE = BY_CODE[0x0000]


class TestCipherSuites(object):

    def test_kx(self):
        # A test from each RFC
        assert (BY_CODE[0x0005].kx == 'RSA')
        assert (BY_CODE[0x0021].kx == 'KRB5')
        assert (BY_CODE[0x002d].kx == 'DHE')
        assert (BY_CODE[0x0034].kx == 'DH')
        assert (BY_CODE[0x003c].kx == 'RSA')
        assert (BY_CODE[0x0042].kx == 'DH')
        assert (BY_CODE[0x006a].kx == 'DHE')
        assert (BY_CODE[0x0084].kx == 'RSA')
        assert (BY_CODE[0x0091].kx == 'DHE')
        assert (BY_CODE[0x0098].kx == 'DH')
        assert (BY_CODE[0x00ab].kx == 'DHE')
        assert (BY_CODE[0x00b0].kx == 'PSK')
        assert (BY_CODE[0x00bb].kx == 'DH')
        assert (BY_CODE[0xc008].kx == 'ECDHE')
        assert (BY_CODE[0xc016].kx == 'ECDH')
        assert (BY_CODE[0xc01d].kx == 'SRP_SHA')
        assert (BY_CODE[0xc027].kx == 'ECDHE')
        assert (BY_CODE[0xc036].kx == 'ECDHE')
        assert (BY_CODE[0xc045].kx == 'DHE')
        assert (BY_CODE[0xc052].kx == 'DHE')
        assert (BY_CODE[0xc068].kx == 'RSA')
        assert (BY_CODE[0xc074].kx == 'ECDH')
        assert (BY_CODE[0xc08d].kx == 'ECDH')
        assert (BY_CODE[0xc09d].kx == 'RSA')
        assert (BY_CODE[0xc0a2].kx == 'DHE')
        assert (BY_CODE[0xc0ad].kx == 'ECDHE')
        assert (BY_CODE[0xcc13].kx == 'ECDHE')
        assert (BY_CODE[0xcca8].kx == 'ECDHE')
        assert (BY_CODE[0xccae].kx == 'RSA')

    def test_auth(self):
        # A test from each RFC
        assert (BY_CODE[0x0005].auth == 'RSA')
        assert (BY_CODE[0x0021].auth == 'KRB5')
        assert (BY_CODE[0x002d].auth == 'PSK')
        assert (BY_CODE[0x0034].auth == 'anon')
        assert (BY_CODE[0x003c].auth == 'RSA')
        assert (BY_CODE[0x0042].auth == 'DSS')
        assert (BY_CODE[0x006a].auth == 'DSS')
        assert (BY_CODE[0x0084].auth == 'RSA')
        assert (BY_CODE[0x0091].auth == 'PSK')
        assert (BY_CODE[0x0098].auth == 'RSA')
        assert (BY_CODE[0x00ab].auth == 'PSK')
        assert (BY_CODE[0x00b0].auth == 'PSK')
        assert (BY_CODE[0x00bb].auth == 'DSS')
        assert (BY_CODE[0xc008].auth == 'ECDSA')
        assert (BY_CODE[0xc016].auth == 'anon')
        assert (BY_CODE[0xc01d].auth == 'SRP_SHA')
        assert (BY_CODE[0xc027].auth == 'RSA')
        assert (BY_CODE[0xc036].auth == 'PSK')
        assert (BY_CODE[0xc045].auth == 'RSA')
        assert (BY_CODE[0xc052].auth == 'RSA')
        assert (BY_CODE[0xc068].auth == 'PSK')
        assert (BY_CODE[0xc074].auth == 'ECDSA')
        assert (BY_CODE[0xc08d].auth == 'RSA')
        assert (BY_CODE[0xc09d].auth == 'RSA')
        assert (BY_CODE[0xc0a2].auth == 'RSA')
        assert (BY_CODE[0xc0ad].auth == 'ECDSA')
        assert (BY_CODE[0xcc14].auth == 'ECDSA')
        assert (BY_CODE[0xcca8].auth == 'RSA')
        assert (BY_CODE[0xccae].auth == 'PSK')

    def test_pfs(self):
        assert (BY_NAME('TLS_RSA_WITH_RC4_128_SHA').pfs is False)
        assert (BY_NAME('TLS_DHE_DSS_WITH_AES_256_CBC_SHA256').pfs is True)
        assert (BY_NAME('TLS_ECDHE_ECDSA_WITH_3DES_EDE_CBC_SHA').pfs is True)

    def test_aead(self):
        assert (BY_NAME('TLS_RSA_WITH_AES_128_CBC_SHA256').aead is False)
        assert (BY_NAME('TLS_RSA_WITH_AES_256_CCM').aead is True)
        assert (BY_NAME('TLS_DHE_RSA_WITH_AES_128_CCM_8').aead is True)
        assert (BY_NAME('TLS_DHE_PSK_WITH_AES_256_GCM_SHA384').aead is True)

    def test_anonymous(self):
        assert (BY_NAME('TLS_RSA_WITH_RC4_128_SHA').anonymous is False)
        assert (BY_NAME('TLS_DH_anon_WITH_AES_128_CBC_SHA').anonymous is True)
        assert (BY_NAME('TLS_DH_anon_EXPORT_WITH_DES40_CBC_SHA').anonymous is True)

    def test_by_name_and_code(self):
        # Special cases:
        # - explicit name
        assert (BY_CODE[0x00ff] == BY_NAME('TLS_EMPTY_RENEGOTIATION_INFO'))
        # - explicit encoding (DES_40 + CBC = DES_CBC_40)
        assert (BY_CODE[0x0026] == BY_NAME('TLS_KRB5_EXPORT_WITH_DES_CBC_40_SHA'))

        # A test from each RFC
        assert (BY_CODE[0x0005] == BY_NAME('TLS_RSA_WITH_RC4_128_SHA'))
        assert (BY_CODE[0x0021] == BY_NAME('TLS_KRB5_WITH_IDEA_CBC_SHA'))
        assert (BY_CODE[0x002d] == BY_NAME('TLS_DHE_PSK_WITH_NULL_SHA'))
        assert (BY_CODE[0x0034] == BY_NAME('TLS_DH_anon_WITH_AES_128_CBC_SHA'))
        assert (BY_CODE[0x003c] == BY_NAME('TLS_RSA_WITH_AES_128_CBC_SHA256'))
        assert (BY_CODE[0x0042] == BY_NAME('TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA'))
        assert (BY_CODE[0x006a] == BY_NAME('TLS_DHE_DSS_WITH_AES_256_CBC_SHA256'))
        assert (BY_CODE[0x0084] == BY_NAME('TLS_RSA_WITH_CAMELLIA_256_CBC_SHA'))
        assert (BY_CODE[0x0091] == BY_NAME('TLS_DHE_PSK_WITH_AES_256_CBC_SHA'))
        assert (BY_CODE[0x0098] == BY_NAME('TLS_DH_RSA_WITH_SEED_CBC_SHA'))
        assert (BY_CODE[0x00ab] == BY_NAME('TLS_DHE_PSK_WITH_AES_256_GCM_SHA384'))
        assert (BY_CODE[0x00b0] == BY_NAME('TLS_PSK_WITH_NULL_SHA256'))
        assert (BY_CODE[0x00bb] == BY_NAME('TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA256'))
        assert (BY_CODE[0x1303] == BY_NAME('TLS_CHACHA20_POLY1305_SHA256'))
        assert (BY_CODE[0xc008] == BY_NAME('TLS_ECDHE_ECDSA_WITH_3DES_EDE_CBC_SHA'))
        assert (BY_CODE[0xc016] == BY_NAME('TLS_ECDH_anon_WITH_RC4_128_SHA'))
        assert (BY_CODE[0xc01d] == BY_NAME('TLS_SRP_SHA_WITH_AES_128_CBC_SHA'))
        assert (BY_CODE[0xc027] == BY_NAME('TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256'))
        assert (BY_CODE[0xc036] == BY_NAME('TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA'))
        assert (BY_CODE[0xc045] == BY_NAME('TLS_DHE_RSA_WITH_ARIA_256_CBC_SHA384'))
        assert (BY_CODE[0xc052] == BY_NAME('TLS_DHE_RSA_WITH_ARIA_128_GCM_SHA256'))
        assert (BY_CODE[0xc068] == BY_NAME('TLS_RSA_PSK_WITH_ARIA_128_CBC_SHA256'))
        assert (BY_CODE[0xc074] == BY_NAME('TLS_ECDH_ECDSA_WITH_CAMELLIA_128_CBC_SHA256'))
        assert (BY_CODE[0xc08d] == BY_NAME('TLS_ECDH_RSA_WITH_CAMELLIA_256_GCM_SHA384'))
        assert (BY_CODE[0xc09d] == BY_NAME('TLS_RSA_WITH_AES_256_CCM'))
        assert (BY_CODE[0xc0a2] == BY_NAME('TLS_DHE_RSA_WITH_AES_128_CCM_8'))
        assert (BY_CODE[0xc0ad] == BY_NAME('TLS_ECDHE_ECDSA_WITH_AES_256_CCM'))
        assert (BY_CODE[0xcca8] == BY_NAME('TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256'))
        assert (BY_CODE[0xccae] == BY_NAME('TLS_RSA_PSK_WITH_CHACHA20_POLY1305_SHA256'))
        assert (BY_CODE[0xcc15] == BY_NAME('OLD_TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256'))

    def test_repr(self):
        cs = CipherSuite(0x0009, 'RSA', '         ', 'DES     ', 'CBC ', 'SHA')
        assert repr(cs) == "CipherSuite(0x0009, TLS_RSA_WITH_DES_CBC_SHA)"

        assert cs.mac_size == 20
        assert cs.block_size == 8

        repr(BY_CODE[0x6a6a]) == "CipherSuite(0x6a6a, GREASE)"



# $Id: 80211.py 53 2008-12-18 01:22:57Z jon.oberheide $
# -*- coding: utf-8 -*-
"""IEEE 802.11."""
from __future__ import print_function
from __future__ import absolute_import

import struct

from . import dpkt
from .compat import ntole, ntole64

# Frame Types
MGMT_TYPE = 0
CTL_TYPE = 1
DATA_TYPE = 2

# Frame Sub-Types
M_ASSOC_REQ = 0
M_ASSOC_RESP = 1
M_REASSOC_REQ = 2
M_REASSOC_RESP = 3
M_PROBE_REQ = 4
M_PROBE_RESP = 5
M_BEACON = 8
M_ATIM = 9
M_DISASSOC = 10
M_AUTH = 11
M_DEAUTH = 12
M_ACTION = 13
C_BLOCK_ACK_REQ = 8
C_BLOCK_ACK = 9
C_PS_POLL = 10
C_RTS = 11
C_CTS = 12
C_ACK = 13
C_CF_END = 14
C_CF_END_ACK = 15
D_DATA = 0
D_DATA_CF_ACK = 1
D_DATA_CF_POLL = 2
D_DATA_CF_ACK_POLL = 3
D_NULL = 4
D_CF_ACK = 5
D_CF_POLL = 6
D_CF_ACK_POLL = 7
D_QOS_DATA = 8
D_QOS_CF_ACK = 9
D_QOS_CF_POLL = 10
D_QOS_CF_ACK_POLL = 11
D_QOS_NULL = 12
D_QOS_CF_POLL_EMPTY = 14

TO_DS_FLAG = 10
FROM_DS_FLAG = 1
INTER_DS_FLAG = 11

# Bitshifts for Frame Control
_VERSION_MASK = 0x0300
_TYPE_MASK = 0x0c00
_SUBTYPE_MASK = 0xf000
_TO_DS_MASK = 0x0001
_FROM_DS_MASK = 0x0002
_MORE_FRAG_MASK = 0x0004
_RETRY_MASK = 0x0008
_PWR_MGT_MASK = 0x0010
_MORE_DATA_MASK = 0x0020
_WEP_MASK = 0x0040
_ORDER_MASK = 0x0080
_FRAGMENT_NUMBER_MASK = 0x000F
_SEQUENCE_NUMBER_MASK = 0XFFF0
_VERSION_SHIFT = 8
_TYPE_SHIFT = 10
_SUBTYPE_SHIFT = 12
_TO_DS_SHIFT = 0
_FROM_DS_SHIFT = 1
_MORE_FRAG_SHIFT = 2
_RETRY_SHIFT = 3
_PWR_MGT_SHIFT = 4
_MORE_DATA_SHIFT = 5
_WEP_SHIFT = 6
_ORDER_SHIFT = 7
_SEQUENCE_NUMBER_SHIFT = 4

# IEs
IE_SSID = 0
IE_RATES = 1
IE_FH = 2
IE_DS = 3
IE_CF = 4
IE_TIM = 5
IE_IBSS = 6
IE_HT_CAPA = 45
IE_ESR = 50
IE_HT_INFO = 61

FCS_LENGTH = 4

FRAMES_WITH_CAPABILITY = [M_BEACON, M_ASSOC_RESP, M_ASSOC_REQ, M_REASSOC_REQ, ]

# Block Ack control constants
_ACK_POLICY_SHIFT = 0
_MULTI_TID_SHIFT = 1
_COMPRESSED_SHIFT = 2
_TID_SHIFT = 12

_ACK_POLICY_MASK = 0x0001
_MULTI_TID_MASK = 0x0002
_COMPRESSED_MASK = 0x0004
_TID_MASK = 0xf000

_COMPRESSED_BMP_LENGTH = 8
_BMP_LENGTH = 128

# Action frame categories
BLOCK_ACK = 3

# Block ack category action codes
BLOCK_ACK_CODE_REQUEST = 0
BLOCK_ACK_CODE_RESPONSE = 1
BLOCK_ACK_CODE_DELBA = 2


class IEEE80211(dpkt.Packet):
    """IEEE 802.11.

    IEEE 802.11 is part of the IEEE 802 set of local area network (LAN) technical standards,
    and specifies the set of media access control (MAC) and physical layer (PHY) protocols
    for implementing wireless local area network (WLAN) computer communication.

    Attributes:
        __hdr__: Header fields of IEEE802.11.
            framectl: (int): Frame control (2 bytes)
            duration: (int): Duration ID (2 bytes)
    """

    __hdr__ = (
        ('framectl', 'H', 0),
        ('duration', 'H', 0)
    )

    # The standard really defines the entire MAC protocol as little-endian on the wire,
    # however there is broken logic in the rest of the module preventing this from working right now
    #  __byte_order__ = '<'

    @property
    def version(self):
        return (self.framectl & _VERSION_MASK) >> _VERSION_SHIFT

    @version.setter
    def version(self, val):
        self.framectl = (val << _VERSION_SHIFT) | (self.framectl & ~_VERSION_MASK)

    @property
    def type(self):
        return (self.framectl & _TYPE_MASK) >> _TYPE_SHIFT

    @type.setter
    def type(self, val):
        self.framectl = (val << _TYPE_SHIFT) | (self.framectl & ~_TYPE_MASK)

    @property
    def subtype(self):
        return (self.framectl & _SUBTYPE_MASK) >> _SUBTYPE_SHIFT

    @subtype.setter
    def subtype(self, val):
        self.framectl = (val << _SUBTYPE_SHIFT) | (self.framectl & ~_SUBTYPE_MASK)

    @property
    def to_ds(self):
        return (self.framectl & _TO_DS_MASK) >> _TO_DS_SHIFT

    @to_ds.setter
    def to_ds(self, val):
        self.framectl = (val << _TO_DS_SHIFT) | (self.framectl & ~_TO_DS_MASK)

    @property
    def from_ds(self):
        return (self.framectl & _FROM_DS_MASK) >> _FROM_DS_SHIFT

    @from_ds.setter
    def from_ds(self, val):
        self.framectl = (val << _FROM_DS_SHIFT) | (self.framectl & ~_FROM_DS_MASK)

    @property
    def more_frag(self):
        return (self.framectl & _MORE_FRAG_MASK) >> _MORE_FRAG_SHIFT

    @more_frag.setter
    def more_frag(self, val):
        self.framectl = (val << _MORE_FRAG_SHIFT) | (self.framectl & ~_MORE_FRAG_MASK)

    @property
    def retry(self):
        return (self.framectl & _RETRY_MASK) >> _RETRY_SHIFT

    @retry.setter
    def retry(self, val):
        self.framectl = (val << _RETRY_SHIFT) | (self.framectl & ~_RETRY_MASK)

    @property
    def pwr_mgt(self):
        return (self.framectl & _PWR_MGT_MASK) >> _PWR_MGT_SHIFT

    @pwr_mgt.setter
    def pwr_mgt(self, val):
        self.framectl = (val << _PWR_MGT_SHIFT) | (self.framectl & ~_PWR_MGT_MASK)

    @property
    def more_data(self):
        return (self.framectl & _MORE_DATA_MASK) >> _MORE_DATA_SHIFT

    @more_data.setter
    def more_data(self, val):
        self.framectl = (val << _MORE_DATA_SHIFT) | (self.framectl & ~_MORE_DATA_MASK)

    @property
    def wep(self):
        return (self.framectl & _WEP_MASK) >> _WEP_SHIFT

    @wep.setter
    def wep(self, val):
        self.framectl = (val << _WEP_SHIFT) | (self.framectl & ~_WEP_MASK)

    @property
    def order(self):
        return (self.framectl & _ORDER_MASK) >> _ORDER_SHIFT

    @order.setter
    def order(self, val):
        self.framectl = (val << _ORDER_SHIFT) | (self.framectl & ~_ORDER_MASK)

    def unpack_ies(self, buf):
        self.ies = []

        ie_decoder = {
            IE_SSID: ('ssid', self.IE),
            IE_RATES: ('rate', self.IE),
            IE_FH: ('fh', self.FH),
            IE_DS: ('ds', self.DS),
            IE_CF: ('cf', self.CF),
            IE_TIM: ('tim', self.TIM),
            IE_IBSS: ('ibss', self.IBSS),
            IE_HT_CAPA: ('ht_capa', self.IE),
            IE_ESR: ('esr', self.IE),
            IE_HT_INFO: ('ht_info', self.IE)
        }

        # each IE starts with an ID and a length
        while len(buf) > FCS_LENGTH:
            ie_id = struct.unpack('B', buf[:1])[0]
            try:
                parser = ie_decoder[ie_id][1]
                name = ie_decoder[ie_id][0]
            except KeyError:
                parser = self.IE
                name = 'ie_' + str(ie_id)
            ie = parser(buf)

            ie.data = buf[2:2 + ie.len]
            setattr(self, name, ie)
            self.ies.append(ie)
            buf = buf[2 + ie.len:]

    class Capability(object):
        def __init__(self, field):
            self.ess = field & 1
            self.ibss = (field >> 1) & 1
            self.cf_poll = (field >> 2) & 1
            self.cf_poll_req = (field >> 3) & 1
            self.privacy = (field >> 4) & 1
            self.short_preamble = (field >> 5) & 1
            self.pbcc = (field >> 6) & 1
            self.hopping = (field >> 7) & 1
            self.spec_mgmt = (field >> 8) & 1
            self.qos = (field >> 9) & 1
            self.short_slot = (field >> 10) & 1
            self.apsd = (field >> 11) & 1
            self.dsss = (field >> 13) & 1
            self.delayed_blk_ack = (field >> 14) & 1
            self.imm_blk_ack = (field >> 15) & 1

    def __init__(self, *args, **kwargs):
        if kwargs and 'fcs' in kwargs:
            self.fcs_present = kwargs.pop('fcs')
        else:
            self.fcs_present = False

        super(IEEE80211, self).__init__(*args, **kwargs)

    def unpack(self, buf):
        dpkt.Packet.unpack(self, buf)
        self.data = buf[self.__hdr_len__:]

        m_decoder = {
            M_BEACON: ('beacon', self.Beacon),
            M_ASSOC_REQ: ('assoc_req', self.Assoc_Req),
            M_ASSOC_RESP: ('assoc_resp', self.Assoc_Resp),
            M_DISASSOC: ('diassoc', self.Disassoc),
            M_REASSOC_REQ: ('reassoc_req', self.Reassoc_Req),
            M_REASSOC_RESP: ('reassoc_resp', self.Assoc_Resp),
            M_AUTH: ('auth', self.Auth),
            M_PROBE_RESP: ('probe_resp', self.Beacon),
            M_DEAUTH: ('deauth', self.Deauth),
            M_ACTION: ('action', self.Action)
        }

        c_decoder = {
            C_RTS: ('rts', self.RTS),
            C_CTS: ('cts', self.CTS),
            C_ACK: ('ack', self.ACK),
            C_BLOCK_ACK_REQ: ('bar', self.BlockAckReq),
            C_BLOCK_ACK: ('back', self.BlockAck),
            C_CF_END: ('cf_end', self.CFEnd),
        }

        d_dsData = {
            0: self.Data,
            FROM_DS_FLAG: self.DataFromDS,
            TO_DS_FLAG: self.DataToDS,
            INTER_DS_FLAG: self.DataInterDS
        }

        # For now decode everything with DATA. Haven't checked about other QoS
        # additions
        d_decoder = {
            # modified the decoder to consider the ToDS and FromDS flags
            # Omitting the 11 case for now
            D_DATA: ('data_frame', d_dsData),
            D_NULL: ('data_frame', d_dsData),
            D_QOS_DATA: ('data_frame', d_dsData),
            D_QOS_NULL: ('data_frame', d_dsData)
        }

        decoder = {
            MGMT_TYPE: m_decoder,
            CTL_TYPE: c_decoder,
            DATA_TYPE: d_decoder
        }

        # Strip off the FCS field
        if self.fcs_present:
            self.fcs = struct.unpack('<I', self.data[-1 * FCS_LENGTH:])[0]
            self.data = self.data[0: -1 * FCS_LENGTH]

        if self.type == MGMT_TYPE:
            self.mgmt = self.MGMT_Frame(self.data)
            self.data = self.mgmt.data
            if self.subtype == M_PROBE_REQ:
                self.unpack_ies(self.data)
                return
            if self.subtype == M_ATIM:
                return

        try:
            parser = decoder[self.type][self.subtype][1]
            name = decoder[self.type][self.subtype][0]
        except KeyError:
            raise dpkt.UnpackError("KeyError: type=%s subtype=%s" % (self.type, self.subtype))

        if self.type == DATA_TYPE:
            # need to grab the ToDS/FromDS info
            parser = parser[self.to_ds * 10 + self.from_ds]

        if self.type == MGMT_TYPE:
            field = parser(self.mgmt.data)
        else:
            field = parser(self.data)
            self.data = field

        setattr(self, name, field)

        if self.type == MGMT_TYPE:
            self.unpack_ies(field.data)
            if self.subtype in FRAMES_WITH_CAPABILITY:
                self.capability = self.Capability(ntole(field.capability))

        if self.type == DATA_TYPE and self.subtype == D_QOS_DATA:
            self.qos_data = self.QoS_Data(field.data)
            field.data = self.qos_data.data

        self.data = field.data

    class BlockAckReq(dpkt.Packet):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
            ('src', '6s', '\x00' * 6),
            ('ctl', 'H', 0),
            ('seq', 'H', 0),
        )

    class BlockAck(dpkt.Packet):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
            ('src', '6s', '\x00' * 6),
            ('ctl', 'H', 0),
            ('seq', 'H', 0),
        )

        @property
        def compressed(self):
            return (self.ctl & _COMPRESSED_MASK) >> _COMPRESSED_SHIFT

        @compressed.setter
        def compressed(self, val):
            self.ctl = (val << _COMPRESSED_SHIFT) | (self.ctl & ~_COMPRESSED_MASK)

        @property
        def ack_policy(self):
            return (self.ctl & _ACK_POLICY_MASK) >> _ACK_POLICY_SHIFT

        @ack_policy.setter
        def ack_policy(self, val):
            self.ctl = (val << _ACK_POLICY_SHIFT) | (self.ctl & ~_ACK_POLICY_MASK)

        @property
        def multi_tid(self):
            return (self.ctl & _MULTI_TID_MASK) >> _MULTI_TID_SHIFT

        @multi_tid.setter
        def multi_tid(self, val):
            self.ctl = (val << _MULTI_TID_SHIFT) | (self.ctl & ~_MULTI_TID_MASK)

        @property
        def tid(self):
            return (self.ctl & _TID_MASK) >> _TID_SHIFT

        @tid.setter
        def tid(self, val):
            self.ctl = (val << _TID_SHIFT) | (self.ctl & ~_TID_MASK)

        def unpack(self, buf):
            dpkt.Packet.unpack(self, buf)
            self.data = buf[self.__hdr_len__:]
            self.ctl = ntole(self.ctl)

            if self.compressed:
                self.bmp = struct.unpack('8s', self.data[0:_COMPRESSED_BMP_LENGTH])[0]
            else:
                self.bmp = struct.unpack('128s', self.data[0:_BMP_LENGTH])[0]
            self.data = self.data[len(self.__hdr__) + len(self.bmp):]

    class _FragmentNumSeqNumMixin(object):
        @property
        def fragment_number(self):
            return ntole(self.frag_seq) & _FRAGMENT_NUMBER_MASK

        @property
        def sequence_number(self):
            return (ntole(self.frag_seq) & _SEQUENCE_NUMBER_MASK) >> _SEQUENCE_NUMBER_SHIFT

    class RTS(dpkt.Packet):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
            ('src', '6s', '\x00' * 6)
        )

    class CTS(dpkt.Packet):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
        )

    class ACK(dpkt.Packet):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
        )

    class CFEnd(dpkt.Packet):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
            ('src', '6s', '\x00' * 6),
        )

    class MGMT_Frame(dpkt.Packet, _FragmentNumSeqNumMixin):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
            ('src', '6s', '\x00' * 6),
            ('bssid', '6s', '\x00' * 6),
            ('frag_seq', 'H', 0)
        )

    class Beacon(dpkt.Packet):
        __hdr__ = (
            ('timestamp', 'Q', 0),
            ('interval', 'H', 0),
            ('capability', 'H', 0)
        )

        def unpack(self, buf):
            dpkt.Packet.unpack(self, buf)
            self.timestamp = ntole64(self.timestamp)
            self.interval = ntole(self.interval)

    class Disassoc(dpkt.Packet):
        __hdr__ = (
            ('reason', 'H', 0),
        )

    class Assoc_Req(dpkt.Packet):
        __hdr__ = (
            ('capability', 'H', 0),
            ('interval', 'H', 0)
        )

    class Assoc_Resp(dpkt.Packet):
        __hdr__ = (
            ('capability', 'H', 0),
            ('status', 'H', 0),
            ('aid', 'H', 0)
        )

    class Reassoc_Req(dpkt.Packet):
        __hdr__ = (
            ('capability', 'H', 0),
            ('interval', 'H', 0),
            ('current_ap', '6s', '\x00' * 6)
        )

    # This obviously doesn't support any of AUTH frames that use encryption
    class Auth(dpkt.Packet):
        __hdr__ = (
            ('algorithm', 'H', 0),
            ('auth_seq', 'H', 0),
        )

    class Deauth(dpkt.Packet):
        __hdr__ = (
            ('reason', 'H', 0),
        )

    class Action(dpkt.Packet):
        __hdr__ = (
            ('category', 'B', 0),
            ('code', 'B', 0),
        )

        def unpack(self, buf):
            dpkt.Packet.unpack(self, buf)

            action_parser = {
                BLOCK_ACK: {
                    BLOCK_ACK_CODE_REQUEST: ('block_ack_request', IEEE80211.BlockAckActionRequest),
                    BLOCK_ACK_CODE_RESPONSE: ('block_ack_response', IEEE80211.BlockAckActionResponse),
                    BLOCK_ACK_CODE_DELBA: ('block_ack_delba', IEEE80211.BlockAckActionDelba),
                },
            }

            try:
                decoder = action_parser[self.category][self.code][1]
                field_name = action_parser[self.category][self.code][0]
            except KeyError:
                raise dpkt.UnpackError("KeyError: category=%s code=%s" % (self.category, self.code))

            field = decoder(self.data)
            setattr(self, field_name, field)
            self.data = field.data

    class BlockAckActionRequest(dpkt.Packet):
        __hdr__ = (
            ('dialog', 'B', 0),
            ('parameters', 'H', 0),
            ('timeout', 'H', 0),
            ('starting_seq', 'H', 0),
        )

    class BlockAckActionResponse(dpkt.Packet):
        __hdr__ = (
            ('dialog', 'B', 0),
            ('status_code', 'H', 0),
            ('parameters', 'H', 0),
            ('timeout', 'H', 0),
        )

    class BlockAckActionDelba(dpkt.Packet):
        __byte_order__ = '<'
        __hdr__ = (
            ('delba_param_set', 'H', 0),
            ('reason_code', 'H', 0),
            # ('gcr_group_addr', '8s', '\x00' * 8), # Standard says it must be there, but it isn't?
        )

    class Data(dpkt.Packet, _FragmentNumSeqNumMixin):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
            ('src', '6s', '\x00' * 6),
            ('bssid', '6s', '\x00' * 6),
            ('frag_seq', 'H', 0)
        )

    class DataFromDS(dpkt.Packet, _FragmentNumSeqNumMixin):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
            ('bssid', '6s', '\x00' * 6),
            ('src', '6s', '\x00' * 6),
            ('frag_seq', 'H', 0)
        )

    class DataToDS(dpkt.Packet, _FragmentNumSeqNumMixin):
        __hdr__ = (
            ('bssid', '6s', '\x00' * 6),
            ('src', '6s', '\x00' * 6),
            ('dst', '6s', '\x00' * 6),
            ('frag_seq', 'H', 0)
        )

    class DataInterDS(dpkt.Packet, _FragmentNumSeqNumMixin):
        __hdr__ = (
            ('dst', '6s', '\x00' * 6),
            ('src', '6s', '\x00' * 6),
            ('da', '6s', '\x00' * 6),
            ('frag_seq', 'H', 0),
            ('sa', '6s', '\x00' * 6)
        )

    class QoS_Data(dpkt.Packet):
        __hdr__ = (
            ('control', 'H', 0),
        )

    class IE(dpkt.Packet):
        __hdr__ = (
            ('id', 'B', 0),
            ('len', 'B', 0)
        )

        def unpack(self, buf):
            dpkt.Packet.unpack(self, buf)
            self.info = buf[2:self.len + 2]

    class FH(dpkt.Packet):
        __hdr__ = (
            ('id', 'B', 0),
            ('len', 'B', 0),
            ('tu', 'H', 0),
            ('hopset', 'B', 0),
            ('hoppattern', 'B', 0),
            ('hopindex', 'B', 0)
        )

    class DS(dpkt.Packet):
        __hdr__ = (
            ('id', 'B', 0),
            ('len', 'B', 0),
            ('ch', 'B', 0)
        )

    class CF(dpkt.Packet):
        __hdr__ = (
            ('id', 'B', 0),
            ('len', 'B', 0),
            ('count', 'B', 0),
            ('period', 'B', 0),
            ('max', 'H', 0),
            ('dur', 'H', 0)
        )

    class TIM(dpkt.Packet):
        __hdr__ = (
            ('id', 'B', 0),
            ('len', 'B', 0),
            ('count', 'B', 0),
            ('period', 'B', 0),
            ('ctrl', 'H', 0)
        )

        def unpack(self, buf):
            dpkt.Packet.unpack(self, buf)
            self.bitmap = buf[5:self.len + 2]

    class IBSS(dpkt.Packet):
        __hdr__ = (
            ('id', 'B', 0),
            ('len', 'B', 0),
            ('atim', 'H', 0)
        )


def test_802211_ack():
    s = b'\xd4\x00\x00\x00\x00\x12\xf0\xb6\x1c\xa4\xff\xff\xff\xff'
    ieee = IEEE80211(s, fcs=True)
    assert ieee.version == 0
    assert ieee.type == CTL_TYPE
    assert ieee.subtype == C_ACK
    assert ieee.to_ds == 0
    assert ieee.from_ds == 0
    assert ieee.pwr_mgt == 0
    assert ieee.more_data == 0
    assert ieee.wep == 0
    assert ieee.order == 0
    assert ieee.ack.dst == b'\x00\x12\xf0\xb6\x1c\xa4'
    fcs = struct.unpack('<I', s[-4:])[0]
    assert ieee.fcs == fcs


def test_80211_beacon():
    s = (
        b'\x80\x00\x00\x00\xff\xff\xff\xff\xff\xff\x00\x26\xcb\x18\x6a\x30\x00\x26\xcb\x18\x6a\x30'
        b'\xa0\xd0\x77\x09\x32\x03\x8f\x00\x00\x00\x66\x00\x31\x04\x00\x04\x43\x41\x45\x4e\x01\x08'
        b'\x82\x84\x8b\x0c\x12\x96\x18\x24\x03\x01\x01\x05\x04\x00\x01\x00\x00\x07\x06\x55\x53\x20'
        b'\x01\x0b\x1a\x0b\x05\x00\x00\x6e\x00\x00\x2a\x01\x02\x2d\x1a\x6e\x18\x1b\xff\xff\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x30\x14\x01'
        b'\x00\x00\x0f\xac\x04\x01\x00\x00\x0f\xac\x04\x01\x00\x00\x0f\xac\x01\x28\x00\x32\x04\x30'
        b'\x48\x60\x6c\x36\x03\x51\x63\x03\x3d\x16\x01\x00\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x85\x1e\x05\x00\x8f\x00\x0f\x00\xff\x03\x59\x00'
        b'\x63\x73\x65\x2d\x33\x39\x31\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x36\x96\x06'
        b'\x00\x40\x96\x00\x14\x00\xdd\x18\x00\x50\xf2\x02\x01\x01\x80\x00\x03\xa4\x00\x00\x27\xa4'
        b'\x00\x00\x42\x43\x5e\x00\x62\x32\x2f\x00\xdd\x06\x00\x40\x96\x01\x01\x04\xdd\x05\x00\x40'
        b'\x96\x03\x05\xdd\x05\x00\x40\x96\x0b\x09\xdd\x08\x00\x40\x96\x13\x01\x00\x34\x01\xdd\x05'
        b'\x00\x40\x96\x14\x05'
    )
    ieee = IEEE80211(s, fcs=True)
    assert ieee.version == 0
    assert ieee.type == MGMT_TYPE
    assert ieee.subtype == M_BEACON
    assert ieee.to_ds == 0
    assert ieee.from_ds == 0
    assert ieee.pwr_mgt == 0
    assert ieee.more_data == 0
    assert ieee.wep == 0
    assert ieee.order == 0
    assert ieee.mgmt.dst == b'\xff\xff\xff\xff\xff\xff'
    assert ieee.mgmt.src == b'\x00\x26\xcb\x18\x6a\x30'
    assert ieee.beacon.capability == 0x3104
    assert ieee.capability.privacy == 1
    assert ieee.ssid.data == b'CAEN'
    assert ieee.rate.data == b'\x82\x84\x8b\x0c\x12\x96\x18\x24'
    assert ieee.ds.data == b'\x01'
    assert ieee.tim.data == b'\x00\x01\x00\x00'
    fcs = struct.unpack('<I', s[-4:])[0]
    assert ieee.fcs == fcs


def test_80211_data():
    s = (
        b'\x08\x09\x20\x00\x00\x26\xcb\x17\x3d\x91\x00\x16\x44\xb0\xae\xc6\x00\x02\xb3\xd6\x26\x3c'
        b'\x80\x7e\xaa\xaa\x03\x00\x00\x00\x08\x00\x45\x00\x00\x28\x07\x27\x40\x00\x80\x06\x1d\x39'
        b'\x8d\xd4\x37\x3d\x3f\xf5\xd1\x69\xc0\x5f\x01\xbb\xb2\xd6\xef\x23\x38\x2b\x4f\x08\x50\x10'
        b'\x42\x04\xac\x17\x00\x00'
    )
    ieee = IEEE80211(s, fcs=True)
    assert ieee.type == DATA_TYPE
    assert ieee.subtype == D_DATA
    assert ieee.data_frame.dst == b'\x00\x02\xb3\xd6\x26\x3c'
    assert ieee.data_frame.src == b'\x00\x16\x44\xb0\xae\xc6'
    assert ieee.data_frame.frag_seq == 0x807e
    assert ieee.data_frame.fragment_number == 0
    assert ieee.data_frame.sequence_number == 2024
    assert ieee.data == (b'\xaa\xaa\x03\x00\x00\x00\x08\x00\x45\x00\x00\x28\x07\x27\x40\x00\x80\x06'
                         b'\x1d\x39\x8d\xd4\x37\x3d\x3f\xf5\xd1\x69\xc0\x5f\x01\xbb\xb2\xd6\xef\x23'
                         b'\x38\x2b\x4f\x08\x50\x10\x42\x04')
    assert ieee.fcs == struct.unpack('<I', b'\xac\x17\x00\x00')[0]

    from . import llc
    llc_pkt = llc.LLC(ieee.data_frame.data)
    ip_pkt = llc_pkt.data
    assert ip_pkt.dst == b'\x3f\xf5\xd1\x69'


def test_80211_data_qos():
    s = (
        b'\x88\x01\x3a\x01\x00\x26\xcb\x17\x44\xf0\x00\x23\xdf\xc9\xc0\x93\x00\x26\xcb\x17\x44\xf0'
        b'\x20\x7b\x00\x00\xaa\xaa\x03\x00\x00\x00\x88\x8e\x01\x00\x00\x74\x02\x02\x00\x74\x19\x80'
        b'\x00\x00\x00\x6a\x16\x03\x01\x00\x65\x01\x00\x00\x61\x03\x01\x4b\x4c\xa7\x7e\x27\x61\x6f'
        b'\x02\x7b\x3c\x72\x39\xe3\x7b\xd7\x43\x59\x91\x7f\xaa\x22\x47\x51\xb6\x88\x9f\x85\x90\x87'
        b'\x5a\xd1\x13\x20\xe0\x07\x00\x00\x68\xbd\xa4\x13\xb0\xd5\x82\x7e\xc7\xfb\xe7\xcc\xab\x6e'
        b'\x5d\x5a\x51\x50\xd4\x45\xc5\xa1\x65\x53\xad\xb5\x88\x5b\x00\x1a\x00\x2f\x00\x05\x00\x04'
        b'\x00\x35\x00\x0a\x00\x09\x00\x03\x00\x08\x00\x33\x00\x39\x00\x16\x00\x15\x00\x14\x01\x00'
        b'\xff\xff\xff\xff'
    )
    ieee = IEEE80211(s, fcs=True)
    assert ieee.type == DATA_TYPE
    assert ieee.subtype == D_QOS_DATA
    assert ieee.data_frame.dst == b'\x00\x26\xcb\x17\x44\xf0'
    assert ieee.data_frame.src == b'\x00\x23\xdf\xc9\xc0\x93'
    assert ieee.data_frame.frag_seq == 0x207b
    assert ieee.data_frame.fragment_number == 0
    assert ieee.data_frame.sequence_number == 1970
    assert ieee.data == (b'\xaa\xaa\x03\x00\x00\x00\x88\x8e\x01\x00\x00\x74\x02\x02\x00\x74\x19\x80'
                         b'\x00\x00\x00\x6a\x16\x03\x01\x00\x65\x01\x00\x00\x61\x03\x01\x4b\x4c\xa7'
                         b'\x7e\x27\x61\x6f\x02\x7b\x3c\x72\x39\xe3\x7b\xd7\x43\x59\x91\x7f\xaa\x22'
                         b'\x47\x51\xb6\x88\x9f\x85\x90\x87\x5a\xd1\x13\x20\xe0\x07\x00\x00\x68\xbd'
                         b'\xa4\x13\xb0\xd5\x82\x7e\xc7\xfb\xe7\xcc\xab\x6e\x5d\x5a\x51\x50\xd4\x45'
                         b'\xc5\xa1\x65\x53\xad\xb5\x88\x5b\x00\x1a\x00\x2f\x00\x05\x00\x04\x00\x35'
                         b'\x00\x0a\x00\x09\x00\x03\x00\x08\x00\x33\x00\x39\x00\x16\x00\x15\x00\x14\x01\x00')
    assert ieee.qos_data.control == 0x0
    assert ieee.fcs == struct.unpack('<I', b'\xff\xff\xff\xff')[0]


def test_bug():
    s = (b'\x88\x41\x2c\x00\x00\x26\xcb\x17\x44\xf0\x00\x1e\x52\x97\x14\x11\x00\x1f\x6d\xe8\x18\x00'
         b'\xd0\x07\x00\x00\x6f\x00\x00\x20\x00\x00\x00\x00')
    ieee = IEEE80211(s)
    assert ieee.wep == 1


def test_data_ds():
    # verifying the ToDS and FromDS fields and that we're getting the
    # correct values

    s = (b'\x08\x03\x00\x00\x01\x0b\x85\x00\x00\x00\x00\x26\xcb\x18\x73\x50\x01\x0b\x85\x00\x00\x00'
         b'\x00\x89\x00\x26\xcb\x18\x73\x50')
    ieee = IEEE80211(s)
    assert ieee.type == DATA_TYPE
    assert ieee.to_ds == 1
    assert ieee.from_ds == 1
    assert ieee.data_frame.sa == b'\x00\x26\xcb\x18\x73\x50'
    assert ieee.data_frame.src == b'\x00\x26\xcb\x18\x73\x50'
    assert ieee.data_frame.dst == b'\x01\x0b\x85\x00\x00\x00'
    assert ieee.data_frame.da == b'\x01\x0b\x85\x00\x00\x00'

    s = (b'\x88\x41\x50\x01\x00\x26\xcb\x17\x48\xc1\x00\x24\x2c\xe7\xfe\x8a\xff\xff\xff\xff\xff\xff'
         b'\x80\xa0\x00\x00\x09\x1a\x00\x20\x00\x00\x00\x00')
    ieee = IEEE80211(s)
    assert ieee.type == DATA_TYPE
    assert ieee.to_ds == 1
    assert ieee.from_ds == 0
    assert ieee.data_frame.bssid == b'\x00\x26\xcb\x17\x48\xc1'
    assert ieee.data_frame.src == b'\x00\x24\x2c\xe7\xfe\x8a'
    assert ieee.data_frame.dst == b'\xff\xff\xff\xff\xff\xff'

    s = b'\x08\x02\x02\x01\x00\x02\x44\xac\x27\x70\x00\x1f\x33\x39\x75\x44\x00\x1f\x33\x39\x75\x44\x90\xa4'
    ieee = IEEE80211(s)
    assert ieee.type == DATA_TYPE
    assert ieee.to_ds == 0
    assert ieee.from_ds == 1
    assert ieee.data_frame.bssid == b'\x00\x1f\x33\x39\x75\x44'
    assert ieee.data_frame.src == b'\x00\x1f\x33\x39\x75\x44'
    assert ieee.data_frame.dst == b'\x00\x02\x44\xac\x27\x70'


def test_compressed_block_ack():
    s = (b'\x94\x00\x00\x00\x34\xc0\x59\xd6\x3f\x62\xb4\x75\x0e\x46\x83\xc1\x05\x50\x80\xee\x03\x00'
         b'\x00\x00\x00\x00\x00\x00\xa2\xe4\x98\x45')
    ieee = IEEE80211(s, fcs=True)
    assert ieee.type == CTL_TYPE
    assert ieee.subtype == C_BLOCK_ACK
    assert ieee.back.dst == b'\x34\xc0\x59\xd6\x3f\x62'
    assert ieee.back.src == b'\xb4\x75\x0e\x46\x83\xc1'
    assert ieee.back.compressed == 1
    assert len(ieee.back.bmp) == 8
    assert ieee.back.ack_policy == 1
    assert ieee.back.tid == 5


def test_action_block_ack_request():
    s = (b'\xd0\x00\x3a\x01\x00\x23\x14\x36\x52\x30\xb4\x75\x0e\x46\x83\xc1\xb4\x75\x0e\x46\x83\xc1'
         b'\x70\x14\x03\x00\x0d\x02\x10\x00\x00\x40\x29\x06\x50\x33\x9e')
    ieee = IEEE80211(s, fcs=True)
    assert ieee.type == MGMT_TYPE
    assert ieee.subtype == M_ACTION
    assert ieee.action.category == BLOCK_ACK
    assert ieee.action.code == BLOCK_ACK_CODE_REQUEST
    assert ieee.action.block_ack_request.timeout == 0
    parameters = struct.unpack('<H', b'\x10\x02')[0]
    assert ieee.action.block_ack_request.parameters == parameters


def test_action_block_ack_response():
    s = (b'\xd0\x00\x3c\x00\xb4\x75\x0e\x46\x83\xc1\x00\x23\x14\x36\x52\x30\xb4\x75\x0e\x46\x83\xc1'
         b'\xd0\x68\x03\x01\x0d\x00\x00\x02\x10\x88\x13\x9f\xc0\x0b\x75')
    ieee = IEEE80211(s, fcs=True)
    assert ieee.type == MGMT_TYPE
    assert ieee.subtype == M_ACTION
    assert ieee.action.category == BLOCK_ACK
    assert ieee.action.code == BLOCK_ACK_CODE_RESPONSE
    timeout = struct.unpack('<H', b'\x13\x88')[0]
    assert ieee.action.block_ack_response.timeout == timeout
    parameters = struct.unpack('<H', b'\x10\x02')[0]
    assert ieee.action.block_ack_response.parameters == parameters


def test_action_block_ack_delete():
    s = (b'\xd0\x00\x2c\x00\x00\xc1\x41\x06\x13\x0d\x6c\xb2\xae\xae\xde\x80\x6c\xb2\xae\xae\xde\x80'
         b'\xa0\x52\x03\x02\x00\x08\x01\x00\x74\x5d\x0a\xc6')
    ieee = IEEE80211(s, fcs=True)
    assert ieee.type == MGMT_TYPE
    assert ieee.subtype == M_ACTION
    assert ieee.action.category == BLOCK_ACK
    assert ieee.action.code == BLOCK_ACK_CODE_DELBA
    assert ieee.action.block_ack_delba.delba_param_set == 0x0800
    assert ieee.action.block_ack_delba.reason_code == 1


def test_ieee80211_properties():
    ieee80211 = IEEE80211()
    assert ieee80211.version == 0
    ieee80211.version = 1
    assert ieee80211.version == 1

    assert ieee80211.type == 0
    ieee80211.type = 1
    assert ieee80211.type == 1

    assert ieee80211.subtype == 0
    ieee80211.subtype = 1
    assert ieee80211.subtype == 1

    assert ieee80211.to_ds == 0
    ieee80211.to_ds = 1
    assert ieee80211.to_ds == 1

    assert ieee80211.from_ds == 0
    ieee80211.from_ds = 1
    assert ieee80211.from_ds == 1

    assert ieee80211.more_frag == 0
    ieee80211.more_frag = 1
    assert ieee80211.more_frag == 1

    assert ieee80211.retry == 0
    ieee80211.retry = 0
    assert ieee80211.retry == 0

    assert ieee80211.pwr_mgt == 0
    ieee80211.pwr_mgt = 0
    assert ieee80211.pwr_mgt == 0

    assert ieee80211.more_data == 0
    ieee80211.more_data = 0
    assert ieee80211.more_data == 0

    assert ieee80211.wep == 0
    ieee80211.wep = 1
    assert ieee80211.wep == 1

    assert ieee80211.order == 0
    ieee80211.order = 1
    assert ieee80211.order == 1


def test_blockack_properties():
    blockack = IEEE80211.BlockAck()
    assert blockack.compressed == 0
    blockack.compressed = 1
    assert blockack.compressed == 1

    assert blockack.ack_policy == 0
    blockack.ack_policy = 1
    assert blockack.ack_policy == 1

    assert blockack.multi_tid == 0
    blockack.multi_tid = 1
    assert blockack.multi_tid == 1

    assert blockack.tid == 0
    blockack.tid = 1
    assert blockack.tid == 1


def test_ieee80211_unpack():
    import pytest
    from binascii import unhexlify

    buf = unhexlify(
        '4000'  # subtype set to M_PROBE_REQ
        '0000'

        # MGMT_Frame
        '000000000000'  # dst
        '000000000000'  # src
        '000000000000'  # bssid
        '0000'          # frag_seq
    )
    ieee80211 = IEEE80211(buf)
    assert ieee80211.ies == []

    buf = unhexlify(
        '9000'  # subtype set to M_ATIM
        '0000'

        # MGMT_Frame
        '000000000000'  # dst
        '000000000000'  # src
        '000000000000'  # bssid
        '0000'          # frag_seq
    )
    ieee80211 = IEEE80211(buf)
    assert not hasattr(ieee80211, 'ies')

    buf = unhexlify(
        '0c00'  # type set to invalid value
        '0000'
    )
    with pytest.raises(dpkt.UnpackError, match="KeyError: type=3 subtype=0"):
        IEEE80211(buf)


def test_blockack_unpack():
    from binascii import unhexlify
    # unpack a non-compressed BlockAck
    buf = unhexlify(
        '000000000000'
        '000000000000'
        '0000'   # compressed flag not set
        '0000'
    ) + b'\xff' * 128

    blockack = IEEE80211.BlockAck(buf)
    assert blockack.bmp == b'\xff' * 128
    assert blockack.data == b''


def test_action_unpack():
    import pytest
    from binascii import unhexlify
    buf = unhexlify(
        '01'  # category
        '00'  # code (non-existent)
    )
    with pytest.raises(dpkt.UnpackError, match="KeyError: category=1 code=0"):
        IEEE80211.Action(buf)


def test_beacon_unpack():
    beacon_payload = b"\xb9\x71\xfa\x45\x52\x02\x00\x00\x64\x00\x11\x04"
    beacon = IEEE80211.Beacon(beacon_payload)
    assert beacon.timestamp == 0x0000025245fa71b9
    assert beacon.interval == 100
    assert beacon.capability == 0x1104


def test_fragment_and_sequence_values():
    from binascii import unhexlify
    for raw_frag_seq, (expected_frag_num, expected_seq_num) in [
        ("0000", (0, 0)),
        ("0F00", (15, 0)),
        ("0111", (1, 272)),
        ("B3FF", (3, 4091))
    ]:
        buf = unhexlify(
            '000000000000'  # dst
            '000000000000'  # src
            '000000000000'  # bssid
            + raw_frag_seq
        )
        data = IEEE80211.Data(buf)
        assert data.fragment_number == expected_frag_num
        assert data.sequence_number == expected_seq_num



# $Id: sccp.py 23 2006-11-08 15:45:33Z dugsong $
# -*- coding: utf-8 -*-
"""Cisco Skinny Client Control Protocol."""
from __future__ import absolute_import

from . import dpkt

KEYPAD_BUTTON = 0x00000003
OFF_HOOK = 0x00000006
ON_HOOK = 0x00000007
OPEN_RECEIVE_CHANNEL_ACK = 0x00000022
START_TONE = 0x00000082
STOP_TONE = 0x00000083
SET_LAMP = 0x00000086
SET_SPEAKER_MODE = 0x00000088
START_MEDIA_TRANSMIT = 0x0000008A
STOP_MEDIA_TRANSMIT = 0x0000008B
CALL_INFO = 0x0000008F
DEFINE_TIME_DATE = 0x00000094
DISPLAY_TEXT = 0x00000099
OPEN_RECEIVE_CHANNEL = 0x00000105
CLOSE_RECEIVE_CHANNEL = 0x00000106
SELECT_SOFTKEYS = 0x00000110
CALL_STATE = 0x00000111
DISPLAY_PROMPT_STATUS = 0x00000112
CLEAR_PROMPT_STATUS = 0x00000113
ACTIVATE_CALL_PLANE = 0x00000116


class ActivateCallPlane(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('line_instance', 'I', 0),
    )


class CallInfo(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('calling_party_name', '40s', b''),
        ('calling_party', '24s', b''),
        ('called_party_name', '40s', b''),
        ('called_party', '24s', b''),
        ('line_instance', 'I', 0),
        ('call_id', 'I', 0),
        ('call_type', 'I', 0),
        ('orig_called_party_name', '40s', b''),
        ('orig_called_party', '24s', b'')
    )


class CallState(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('call_state', 'I', 12),  # 12: Proceed, 15: Connected
        ('line_instance', 'I', 1),
        ('call_id', 'I', 0)
    )


class ClearPromptStatus(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('line_instance', 'I', 1),
        ('call_id', 'I', 0)
    )


class CloseReceiveChannel(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('conference_id', 'I', 0),
        ('passthruparty_id', 'I', 0),
    )


class DisplayPromptStatus(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('msg_timeout', 'I', 0),
        ('display_msg', '32s', b''),
        ('line_instance', 'I', 1),
        ('call_id', 'I', 0)
    )


class DisplayText(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('display_msg', '36s', b''),
    )


class KeypadButton(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('button', 'I', 0),
    )


class OpenReceiveChannel(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('conference_id', 'I', 0),
        ('passthruparty_id', 'I', 0),
        ('ms_packet', 'I', 0),
        ('payload_capability', 'I', 4),  # 4: G.711 u-law 64k
        ('echo_cancel_type', 'I', 4),
        ('g723_bitrate', 'I', 0),
    )


class OpenReceiveChannelAck(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('channel_status', 'I', 0),
        ('ip', '4s', b''),
        ('port', 'I', 0),
        ('passthruparty_id', 'I', 0),
    )


class SelectStartKeys(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('line_id', 'I', 1),
        ('call_id', 'I', 0),
        ('softkey_set', 'I', 8),
        ('softkey_map', 'I', 0xffffffff)
    )


class SetLamp(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('stimulus', 'I', 9),  # 9: Line
        ('stimulus_instance', 'I', 1),
        ('lamp_mode', 'I', 1),
    )


class SetSpeakerMode(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('speaker', 'I', 2),  # 2: SpeakerOff
    )


class StartMediaTransmission(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('conference_id', 'I', 0),
        ('passthruparty_id', 'I', 0),
        ('ipv4_or_ipv6', 'I', 0),
        ('remote_ip', '16s', b''),
        ('remote_port', 'I', 0),
        ('ms_packet', 'I', 0),
        ('payload_capability', 'I', 4),  # 4: G.711 u-law 64k
        ('precedence', 'I', 0),
        ('silence_suppression', 'I', 0),
        ('max_frames_per_pkt', 'I', 1),
        ('g723_bitrate', 'I', 0),
        ('call_reference', 'I', 0)
    )


class StartTone(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('tone', 'I', 0x24),  # 0x24: AlertingTone
    )


class StopMediaTransmission(dpkt.Packet):
    __byte_order__ = '<'
    __hdr__ = (
        ('conference_id', 'I', 0),
        ('passthruparty_id', 'I', 0),
    )


class SCCP(dpkt.Packet):
    """Cisco Skinny Client Control Protocol.

    TODO: Longer class information....

    Attributes:
        __hdr__: Header fields of SCCP.
        TODO.
    """

    __byte_order__ = '<'
    __hdr__ = (
        ('len', 'I', 0),
        ('rsvd', 'I', 0),
        ('msgid', 'I', 0),
        ('msg', '0s', b''),
    )
    _msgsw = {
        KEYPAD_BUTTON: KeypadButton,
        OPEN_RECEIVE_CHANNEL_ACK: OpenReceiveChannelAck,
        START_TONE: StartTone,
        SET_LAMP: SetLamp,
        START_MEDIA_TRANSMIT: StartMediaTransmission,
        STOP_MEDIA_TRANSMIT: StopMediaTransmission,
        CALL_INFO: CallInfo,
        DISPLAY_TEXT: DisplayText,
        OPEN_RECEIVE_CHANNEL: OpenReceiveChannel,
        CLOSE_RECEIVE_CHANNEL: CloseReceiveChannel,
        CALL_STATE: CallState,
        DISPLAY_PROMPT_STATUS: DisplayPromptStatus,
        CLEAR_PROMPT_STATUS: ClearPromptStatus,
        ACTIVATE_CALL_PLANE: ActivateCallPlane,
    }

    def unpack(self, buf):
        dpkt.Packet.unpack(self, buf)
        n = self.len - 4
        if n > len(self.data):
            raise dpkt.NeedData('not enough data')
        self.msg, self.data = self.data[:n], self.data[n:]
        try:
            p = self._msgsw[self.msgid](self.msg)
            setattr(self, p.__class__.__name__.lower(), p)
        except (KeyError, dpkt.UnpackError):
            pass


def test_sccp():
    import pytest
    from binascii import unhexlify
    buf = unhexlify(
        '08000000'  # len
        '00000000'  # rsvd
        '03000000'  # msgid (KEYPAD_BUTTON)

        'abcdef01'  # msg
        '23456789'  # daat
    )
    sccp = SCCP(buf)
    assert sccp.msg == b'\xab\xcd\xef\x01'
    assert sccp.data == b'\x23\x45\x67\x89'
    assert isinstance(sccp.keypadbutton, KeypadButton)

    # len is too long for data, raises NeedData
    buf = unhexlify(
        '88880000'  # len
        '00000000'  # rsvd
        '00000003'  # msgid (KEYPAD_BUTTON)

        'abcdef01'  # msg
    )
    with pytest.raises(dpkt.NeedData):
        SCCP(buf)

    # msgid is invalid, raises KeyError on _msgsw (silently caught)
    buf = unhexlify(
        '08000000'  # len
        '00000000'  # rsvd
        '00000003'  # msgid (invalid)

        'abcdef01'  # msg
    )
    sccp = SCCP(buf)
    assert sccp.msg == b'\xab\xcd\xef\x01'
    assert sccp.data == b''



# $Id: pim.py 23 2006-11-08 15:45:33Z dugsong $
# -*- coding: utf-8 -*-
"""Protocol Independent Multicast."""
from __future__ import absolute_import

from . import dpkt


class PIM(dpkt.Packet):
    """Protocol Independent Multicast.

    Protocol Independent Multicast (PIM) is a collection of multicast routing protocols, each optimized for a different
    environment. There are two main PIM protocols, PIM Sparse Mode and PIM Dense Mode. A third PIM protocol,
    Bi-directional PIM, is less widely used.

    Attributes:
        __hdr__: Header fields of PIM.
            _v_type: (int): Version (4 bits) and type (4 bits). PIM version number and Message type. (1 byte)
            _rsvd: (int): Reserved. Always cleared to zero. (1 byte)
            sum: (int): Checksum. The 16-bit one's complement of the one's complement sum of the entire PIM message,
                excluding the data portion in the Register message.(2 bytes)
    """

    __hdr__ = (
        ('_v_type', 'B', 0x20),
        ('_rsvd', 'B', 0),
        ('sum', 'H', 0)
    )
    __bit_fields__ = {
        '_v_type': (
            ('v', 4),
            ('type', 4),
        )
    }

    def __bytes__(self):
        if not self.sum:
            self.sum = dpkt.in_cksum(dpkt.Packet.__bytes__(self))
        return dpkt.Packet.__bytes__(self)


def test_pim():
    from binascii import unhexlify
    buf = unhexlify(
        '20'            # _v_type
        '00'            # rsvd
        'df93'          # sum

        '000100020069'  # data
    )
    pimdata = PIM(buf)
    assert bytes(pimdata) == buf
    # force checksum recalculation
    pimdata = PIM(buf)
    pimdata.sum = 0
    assert pimdata.sum == 0
    assert bytes(pimdata) == buf

    assert pimdata.v == 2
    assert pimdata.type == 0

    # test setters
    buf_modified = unhexlify(
        '31'            # _v_type
        '00'            # rsvd
        'df93'          # sum

        '000100020069'  # data
    )
    pimdata.v = 3
    pimdata.type = 1
    assert bytes(pimdata) == buf_modified



# $Id: sctp.py 23 2006-11-08 15:45:33Z dugsong $
# -*- coding: utf-8 -*-
"""Stream Control Transmission Protocol."""
from __future__ import print_function
from __future__ import absolute_import

from . import dpkt
from . import crc32c

# Stream Control Transmission Protocol
# http://tools.ietf.org/html/rfc2960

# Chunk Types
DATA = 0
INIT = 1
INIT_ACK = 2
SACK = 3
HEARTBEAT = 4
HEARTBEAT_ACK = 5
ABORT = 6
SHUTDOWN = 7
SHUTDOWN_ACK = 8
ERROR = 9
COOKIE_ECHO = 10
COOKIE_ACK = 11
ECNE = 12
CWR = 13
SHUTDOWN_COMPLETE = 14


class SCTP(dpkt.Packet):
    """Stream Control Transmission Protocol.

    TODO: Longer class information....

    Attributes:
        __hdr__: Header fields of SCTP.
        TODO.
    """

    __hdr__ = (
        ('sport', 'H', 0),
        ('dport', 'H', 0),
        ('vtag', 'I', 0),
        ('sum', 'I', 0)
    )

    def unpack(self, buf):
        dpkt.Packet.unpack(self, buf)
        l_ = []
        while self.data:
            chunk = Chunk(self.data)
            l_.append(chunk)
            if len(chunk) == 0:
                self.data = b''
                break
            self.data = self.data[len(chunk):]
        self.chunks = l_

    def __len__(self):
        return self.__hdr_len__ + sum(len(x) for x in self.chunks)

    def __bytes__(self):
        l_ = [bytes(x) for x in self.chunks]
        if self.sum == 0:
            s = crc32c.add(0xffffffff, self.pack_hdr())
            for x in l_:
                s = crc32c.add(s, x)
            self.sum = crc32c.done(s)
        return self.pack_hdr() + b''.join(l_)


class Chunk(dpkt.Packet):
    __hdr__ = (
        ('type', 'B', INIT),
        ('flags', 'B', 0),
        ('len', 'H', 0)
    )

    def unpack(self, buf):
        dpkt.Packet.unpack(self, buf)

        self.data = self.data[:self.len - self.__hdr_len__]
        self.padding = b''  # optional padding for DATA chunks

        # SCTP DATA Chunked is padded, 4-bytes aligned
        if self.type == DATA and self.len % 4:
            plen = 4 - self.len % 4  # padded length
            if plen:
                pos = self.__hdr_len__ + len(self.data)  # end of data in buf
                self.padding = buf[pos:pos + plen]

    def __len__(self):
        return self.len + len(self.padding)

    def __bytes__(self):
        return self.pack_hdr() + bytes(self.data) + self.padding


__s = (b'\x80\x44\x00\x50\x00\x00\x00\x00\x30\xba\xef\x54\x01\x00\x00\x3c\x3b\xb9\x9c\x46\x00\x01'
       b'\xa0\x00\x00\x0a\xff\xff\x2b\x2d\x7e\xb2\x00\x05\x00\x08\x9b\xe6\x18\x9b\x00\x05\x00\x08'
       b'\x9b\xe6\x18\x9c\x00\x0c\x00\x06\x00\x05\x00\x00\x80\x00\x00\x04\xc0\x00\x00\x04\xc0\x06'
       b'\x00\x08\x00\x00\x00\x00')


def test_sctp_pack():
    sctp = SCTP(__s)
    assert (__s == bytes(sctp))
    sctp.sum = 0
    assert (__s == bytes(sctp))


def test_sctp_unpack():
    sctp = SCTP(__s)
    assert (sctp.sport == 32836)
    assert (sctp.dport == 80)
    assert (len(sctp.chunks) == 1)
    assert (len(sctp) == 72)
    chunk = sctp.chunks[0]
    assert (chunk.type == INIT)
    assert (chunk.len == 60)


def test_sctp_data_chunk():  # https://github.com/kbandla/dpkt/issues/499
    # packet 5 from 'sctp-www.cap' downloaded from https://wiki.wireshark.org/SampleCaptures
    # chunk len == 419 so requires padding to a 4-byte boundary
    d = (b'\x80\x44\x00\x50\xd2\x6a\xc1\xe5\x70\xe5\x5b\x4c\x00\x03\x01\xa3\x2b\x2d\x7e\xb2\x00\x00'
         b'\x00\x00\x00\x00\x00\x00\x47\x45\x54\x20\x2f\x20\x48\x54\x54\x50\x2f\x31\x2e\x31\x0d\x0a'
         b'\x48\x6f\x73\x74\x3a\x20\x32\x30\x33\x2e\x32\x35\x35\x2e\x32\x35\x32\x2e\x31\x39\x34\x0d'
         b'\x0a\x55\x73\x65\x72\x2d\x41\x67\x65\x6e\x74\x3a\x20\x4d\x6f\x7a\x69\x6c\x6c\x61\x2f\x35'
         b'\x2e\x30\x20\x28\x58\x31\x31\x3b\x20\x55\x3b\x20\x4c\x69\x6e\x75\x78\x20\x69\x36\x38\x36'
         b'\x3b\x20\x6b\x6f\x2d\x4b\x52\x3b\x20\x72\x76\x3a\x31\x2e\x37\x2e\x31\x32\x29\x20\x47\x65'
         b'\x63\x6b\x6f\x2f\x32\x30\x30\x35\x31\x30\x30\x37\x20\x44\x65\x62\x69\x61\x6e\x2f\x31\x2e'
         b'\x37\x2e\x31\x32\x2d\x31\x0d\x0a\x41\x63\x63\x65\x70\x74\x3a\x20\x74\x65\x78\x74\x2f\x78'
         b'\x6d\x6c\x2c\x61\x70\x70\x6c\x69\x63\x61\x74\x69\x6f\x6e\x2f\x78\x6d\x6c\x2c\x61\x70\x70'
         b'\x6c\x69\x63\x61\x74\x69\x6f\x6e\x2f\x78\x68\x74\x6d\x6c\x2b\x78\x6d\x6c\x2c\x74\x65\x78'
         b'\x74\x2f\x68\x74\x6d\x6c\x3b\x71\x3d\x30\x2e\x39\x2c\x74\x65\x78\x74\x2f\x70\x6c\x61\x69'
         b'\x6e\x3b\x71\x3d\x30\x2e\x38\x2c\x69\x6d\x61\x67\x65\x2f\x70\x6e\x67\x2c\x2a\x2f\x2a\x3b'
         b'\x71\x3d\x30\x2e\x35\x0d\x0a\x41\x63\x63\x65\x70\x74\x2d\x4c\x61\x6e\x67\x75\x61\x67\x65'
         b'\x3a\x20\x6b\x6f\x2c\x65\x6e\x2d\x75\x73\x3b\x71\x3d\x30\x2e\x37\x2c\x65\x6e\x3b\x71\x3d'
         b'\x30\x2e\x33\x0d\x0a\x41\x63\x63\x65\x70\x74\x2d\x45\x6e\x63\x6f\x64\x69\x6e\x67\x3a\x20'
         b'\x67\x7a\x69\x70\x2c\x64\x65\x66\x6c\x61\x74\x65\x0d\x0a\x41\x63\x63\x65\x70\x74\x2d\x43'
         b'\x68\x61\x72\x73\x65\x74\x3a\x20\x45\x55\x43\x2d\x4b\x52\x2c\x75\x74\x66\x2d\x38\x3b\x71'
         b'\x3d\x30\x2e\x37\x2c\x2a\x3b\x71\x3d\x30\x2e\x37\x0d\x0a\x4b\x65\x65\x70\x2d\x41\x6c\x69'
         b'\x76\x65\x3a\x20\x33\x30\x30\x0d\x0a\x43\x6f\x6e\x6e\x65\x63\x74\x69\x6f\x6e\x3a\x20\x6b'
         b'\x65\x65\x70\x2d\x61\x6c\x69\x76\x65\x0d\x0a\x0d\x0a\x00')  # <-- ends with \x00 padding

    sctp = SCTP(d)
    assert sctp.chunks
    assert len(sctp.chunks) == 1

    ch = sctp.chunks[0]
    assert ch.type == DATA
    assert ch.len == 419
    assert len(ch) == 420  # 419 +1 byte padding
    assert ch.data[-14:] == b'keep-alive\r\n\r\n'  # no padding byte at the end

    # no remaining sctp data
    assert sctp.data == b''

    # test packing of the padded chunk
    assert bytes(ch) == d[SCTP.__hdr_len__:]
  


def test_malformed_sctp_data_chunk():  
    # packet 7964 from '4.pcap' downloaded from https://research.unsw.edu.au/projects/unsw-nb15-dataset
    d = (b'\x27\x0f\xe1\xc3\xc2\x73\x4d\x32\x4f\x54\x27\x8c' #header
         b'\x0b\x00\x00\x04' #chunk 0, COOKIE_ACK chunk
         b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00') #chunk 1, malformed DATA chunk, size labeled as 0


    sctp = SCTP(d)
    assert sctp.chunks
    assert len(sctp.chunks) == 2

    ch = sctp.chunks[1]
    assert ch.type == DATA
    assert ch.len == 0
    assert len(ch) == 0 
    assert ch.data == b'\x00\x00'
    
    # no remaining sctp data
    assert sctp.data == b''
